#ifndef SysMyProject_Kernel_H
#define SysMyProject_Kernel_H
#include "spmv/texture.h"

 /*!optimized inner product.
 * Kerner function calculating the inner product of a and b using techniques 
 * presented in the presentation "Optimizing CUDA" by Mark Harris.
 * It (theoretically) performes optimal if the number of operations of each thread is approximately O(log(N)),
 * but we would typically have a little more operations per thread to hide memory latency.
 * We therefore would prefer the number of threads in our grid to be smaller than log(N).
 * Assumes that the size of a and b is a multiple of two times the blockSize.
 * 
 * @param	blockSize	template argument giving the blocksize, to optimize the unrolling of our loop.
 * @param	a,b			the vectors to take the inner product of.
 * @param	N			the number of elements in the vectors a and b.
 */
 
template <unsigned int blockSize, class T, bool useCache>
__global__ void vectorVector_kernel(const T *a, const T *b, const size_t N, T *result){
	extern __shared__ T sdata[];
	int tid = threadIdx.x;
	int i= blockIdx.x*(blockSize) + threadIdx.x;
	int gridSize = gridDim.x*(blockSize);
	//load the data from global to shared memory.
	//a number of loads are performed to hide latency.
	sdata[tid] = 0.0f;
	do{
		sdata[tid] += fetch_x<useCache>(i, a) * fetch_y<useCache>(i, b);
		i += gridSize;
	}while(i < N);
	__syncthreads();
	if (blockSize >= 512){ if (tid < 256) sdata[tid] += sdata[tid + 256]; __syncthreads(); }
	if (blockSize >= 256){ if (tid < 128) sdata[tid] += sdata[tid + 128]; __syncthreads(); }
	if (blockSize >= 128){ if (tid <  64) sdata[tid] += sdata[tid +  64]; __syncthreads(); }
	//all operations within a warp performes in parallell (SIMD), and doesn't need __syncthreads, but this is needed in emulation-mode
	#ifdef __DEVICE_EMULATION__
		if (tid <  32){
			if (blockSize >= 64){ sdata[tid] += sdata[tid +  32]; __syncthreads();}
			if (blockSize >= 32){ sdata[tid] += sdata[tid +  16]; __syncthreads();}
			if (blockSize >= 16){ sdata[tid] += sdata[tid +   8]; __syncthreads();}
			if (blockSize >=  8){ sdata[tid] += sdata[tid +   4]; __syncthreads();}
			if (blockSize >=  4){ sdata[tid] += sdata[tid +   2]; __syncthreads();}
			if (blockSize >=  2){ sdata[tid] += sdata[tid +   1]; __syncthreads();}
		}
	#else
		if (tid <  32){
			if (blockSize >= 64){ sdata[tid] += sdata[tid +  32];}
			if (blockSize >= 32){ sdata[tid] += sdata[tid +  16];}
			if (blockSize >= 16){ sdata[tid] += sdata[tid +   8];}
			if (blockSize >=  8){ sdata[tid] += sdata[tid +   4];}
			if (blockSize >=  4){ sdata[tid] += sdata[tid +   2];}
			if (blockSize >=  2){ sdata[tid] += sdata[tid +   1];}
		}
	#endif
	if (tid == 0) result[blockIdx.x] = sdata[0];
}

/* Optimized sum
 * sums the elements of an array in the same manner as innerProduct.
 */
template <unsigned int blockSize, class T, bool useCache>
__global__ void sum_kernel(const T *x, const unsigned int N, T *outdata){
	extern __shared__ T sdata[];
	int tid = threadIdx.x;
	int i = blockIdx.x*(blockSize) + threadIdx.x;
	int gridSize = blockSize*gridDim.x;
	sdata[tid] = (T)0.0;
	do{
		sdata[tid] += fetch_x<useCache>(i, x);
		i += gridSize;
	}while(i < N);
	__syncthreads();
	if (blockSize >= 512){ if (tid < 256) sdata[tid] += sdata[tid + 256]; __syncthreads(); }
	if (blockSize >= 256){ if (tid < 128) sdata[tid] += sdata[tid + 128]; __syncthreads(); }
	if (blockSize >= 128){ if (tid <  64) sdata[tid] += sdata[tid +  64]; __syncthreads(); }
	if (tid <  32){
		if (blockSize >= 64){ sdata[tid] += sdata[tid +  32];}
		if (blockSize >= 32){ sdata[tid] += sdata[tid +  16];}
		if (blockSize >= 16){ sdata[tid] += sdata[tid +   8];}
		if (blockSize >=  8){ sdata[tid] += sdata[tid +   4];}
		if (blockSize >=  4){ sdata[tid] += sdata[tid +   2];}
		if (blockSize >=  2){ sdata[tid] += sdata[tid +   1];}
	}
	if (tid == 0) outdata[blockIdx.x] = sdata[0];
}

/* simple saxpy calculating the result y = ax + y */
template <class T, bool useCache>
__global__ void axpy_kernel(const T a, const T *x, const size_t N, T *y){
	int i = blockIdx.x*blockDim.x + threadIdx.x;
	if (i < N) y[i] = a*fetch_x<useCache>(i, x) + y[i];
}

/* copy a vector of different format */
template <class T, class T2, bool useCache>
__global__ void copy_kernel(const T *x, const size_t N, T2 *y){
	int i = blockIdx.x*blockDim.x + threadIdx.x;
	if (i < N) y[i] = (T2) fetch_x<useCache>(i, x);
}

/* simple scaling of a vector */
template <class T>
__global__ void scale_kernel(const T a, const size_t N, T *v){
	int i = blockIdx.x*blockDim.x + threadIdx.x;
	if (i < N) v[i] = a*v[i];
}

/* simple scaling of a vector */
template <class T>
__global__ void resetToZero_kernel(const size_t N, T *v){
	int i = blockIdx.x*blockDim.x + threadIdx.x;
	if (i < N) v[i] = (T)0.0;
}

template <class T, const unsigned int numberOfDiags>
__global__ void DM_matrixVector_kernel(const T* x, const T* diags, const int* indexes, const size_t N, T* Ax){
	extern __shared__ T sdata[];
    //l is the line index in the diagonal vector we are modifying.
    int tid = threadIdx.x;
    int l = blockIdx.x*blockDim.x + threadIdx.x;
	int ci;
	if (l >= N) return;
	sdata[tid] = 0.0f;
	//ci is the current y-coordinate in the matrix our GpuDM object represents.
	//cannot parallelize more in danger of reading before another thread reads,
	//then write new value after the other thread has read a wrong value.
	#pragma unroll
	for (int i=0; i<numberOfDiags; i++){
		ci = l + indexes[i];
		if ((0 <= ci) && (ci < N)) sdata[tid] += diags[numberOfDiags*ci+i] * x[ci];
	}
	Ax[l] = sdata[tid];
}

#endif	//SysMyProject_Kernel_H
