#ifndef SysCPUCSR_H_
#define SysCPUCSR_H_
#include "CGsystem.h"
#include "commonHeader.h"
#include "spmv/sparse_formats.h"
#include "spmv/spmv_csr_scalar_device.cu.h"
#include "spmv/spmv_csr_vector_device.cu.h"
#include <MyProject_kernel.cu>

#include "cutil_inline.h"

template <typename indexType, typename valueType>
class SysCPUCSR : public CGsystem<valueType>
{
public:
	SysCPUCSR(const csr_matrix<indexType, valueType> csr_h, const valueType *b);
	virtual ~SysCPUCSR();
	virtual valueType ** initVectors(valueType **x, valueType **b, const int length);
	virtual valueType vectorVectorProduct(const valueType*v1, const valueType *v2) const;
	virtual void matrixVectorProduct(const valueType *v, valueType *result) const;
	virtual void axpy(const valueType a, const valueType *x, valueType *y) const;
	virtual void scale(const valueType scalar, valueType *v) const;
	virtual void copy(const valueType *src, valueType **dst) const {for (int i=0; i < csr.num_rows; i++) (*dst)[i] = src[i];}
	virtual void getSolution(valueType **solution) const;
	virtual size_type getN() const {return csr.num_cols;};
	virtual void printVector(const valueType *v, const int numbersPerLine) const;
	virtual void printMatrix() const;
private:
	//host variables:
	int vecArrayLength;
	//hybrid variable:
	csr_matrix<indexType, valueType> csr;
	//device variables:
	valueType *x;
	valueType *b;
	valueType **vecArray;
};

template <typename indexType, typename valueType>
SysCPUCSR<indexType,  valueType>::SysCPUCSR(const csr_matrix<indexType, valueType> csr_h, const valueType *b_h)
{
	
	writeToScreen("Using sparse matrix in the CSR format on the CPU.\n");

    csr.num_rows = csr_h.num_rows;
    csr.num_cols = csr_h.num_cols;
    csr.num_nonzeros = csr_h.num_nonzeros;

	csr.Ap = copy_array_on_host<indexType>(csr_h.Ap, csr.num_cols + 1);
	csr.Aj = copy_array_on_host<indexType>(csr_h.Aj, csr.num_nonzeros);
	csr.Ax = copy_array_on_host<valueType>(csr_h.Ax, csr.num_nonzeros);
	
	x = new_host_array<valueType>(csr.num_rows);
	
	b = copy_array_on_host<valueType>(b_h, csr.num_rows);
}

template <typename indexType, typename valueType>
SysCPUCSR<indexType, valueType>::~SysCPUCSR()
{
	delete_host_array(x);
	delete_host_array(b);
	for (int i=0; i < vecArrayLength; ++i) delete_host_array(vecArray[i]);
	delete_host_matrix(csr);
	free(vecArray);
}

template <typename indexType, typename valueType>
valueType ** SysCPUCSR<indexType,  valueType>::initVectors(valueType **x_e, valueType **b_e, const int length)
{
	*b_e = b;
	*x_e = x;
	vecArrayLength = length;
	vecArray = (valueType **) malloc(vecArrayLength * sizeof(valueType*));
	for (int i=0; i < vecArrayLength; ++i){
		vecArray[i] = new_host_array<valueType>(csr.num_cols);
	}
	return vecArray;
}

template <typename indexType, typename valueType>
valueType SysCPUCSR<indexType,  valueType>::vectorVectorProduct(const valueType *v1, const valueType *v2) const
{
	valueType sum = 0.0;
	for (int i=0; i < csr.num_rows; ++i){
		sum += v1[i] * v2[i];
	}
	return sum;
}

template <typename indexType, typename valueType>
void SysCPUCSR<indexType,  valueType>::matrixVectorProduct(const valueType *v, valueType *result) const
{
	indexType row_start;
	indexType row_end;
	for (indexType row = 0; row < csr.num_rows; row++){
		result[row] = 0.0;
		row_start = csr.Ap[row];
		row_end = csr.Ap[row + 1];
		for (indexType j = row_start; j < row_end; j++){
			result[row] += csr.Ax[j] * v[csr.Aj[j]];
		}
	}
}

template <typename indexType, typename valueType>
void SysCPUCSR<indexType,  valueType>::axpy(const valueType a, const valueType *x_input, valueType *y) const
{
	for (int i=0; i < csr.num_rows; i++){
		y[i] = a*x_input[i] + y[i];
	}
}

template <typename indexType, typename valueType>
void SysCPUCSR<indexType,  valueType>::scale(const valueType scalar, valueType *v) const
{
	if (scalar == (valueType) 0.0){
		for (int i=0; i < csr.num_rows; i++) v[i] = 0.0;
	} else{
		for (int i=0; i < csr.num_rows; i++){
			v[i] *= scalar;
		}
	}
}

template <typename indexType, typename valueType>
void SysCPUCSR<indexType,  valueType>::getSolution(valueType **solution) const
{
	for (int i=0; i < csr.num_rows; i++){
		(*solution)[i] = x[i];
	}
}

template <typename indexType, typename valueType>
void SysCPUCSR<indexType,  valueType>::printVector(const valueType *v, const int numbersPerLine) const
{
	for (unsigned int line=0; line < csr.num_rows/numbersPerLine; line++){
		for (unsigned int row=0; row< numbersPerLine; row++){
			writeToScreen1("%g ", v[line*numbersPerLine + row]);
		}
		writeToScreen("\n");
	}
	writeToScreen("\n");
}

template <typename indexType, typename valueType>
void SysCPUCSR<indexType, valueType>::printMatrix() const
{
	writeToScreen("i \tpr \tjr \tic\n");
	for (indexType i = 0; i < csr.num_nonzeros; ++i){
		writeToScreen1("%i\t",(int) i);
		writeToScreen1("%g\t", csr.Ax[i]);
		writeToScreen1("%i\t", (int)csr.Aj[i]);
		if (i <= csr.num_rows) writeToScreen1("%i", (int)csr.Ap[i]);
		writeToScreen("\n");
	}
}

#endif /*SysCPUCSR_H_*/
