#ifndef SysCPUHYB_H_
#define SysCPUHYB_H_
#include "CGsystem.h"
#include "commonHeader.h"
#include "spmv/sparse_formats.h"
#include "spmv/sparse_conversions.h"
#include "spmv/spmv_csr_scalar_device.cu.h"
#include "spmv/spmv_csr_vector_device.cu.h"
#include <MyProject_kernel.cu>

#include "cutil_inline.h"

template <typename indexType, typename valueType>
class SysCPUHYB : public CGsystem<valueType>
{
public:
	SysCPUHYB(const csr_matrix<indexType, valueType> csr_h, const valueType *b);
	virtual ~SysCPUHYB();
	virtual valueType ** initVectors(valueType **x, valueType **b, const int length);
	virtual valueType vectorVectorProduct(const valueType*v1, const valueType *v2) const;
	virtual void matrixVectorProduct(const valueType *v, valueType *result) const;
	virtual void axpy(const valueType a, const valueType *x, valueType *y) const;
	virtual void scale(const valueType scalar, valueType *v) const;
	virtual void copy(const valueType *src, valueType **dst) const {for (int i=0; i < hyb.num_rows; i++) (*dst)[i] = src[i];}
	virtual void getSolution(valueType **solution) const;
	virtual size_type getN() const {return hyb.num_cols;};
	virtual void printVector(const valueType *v, const int numbersPerLine) const;
	virtual void printMatrix() const;
private:
	//host variables:
	int vecArrayLength;
	//hybrid variable:
	hyb_matrix<indexType, valueType> hyb;
	//device variables:
	valueType *x;
	valueType *b;
	valueType **vecArray;
};

template <typename indexType, typename valueType>
SysCPUHYB<indexType,  valueType>::SysCPUHYB(const csr_matrix<indexType, valueType> csr_h, const valueType *b_h)
{
	
	writeToScreen("Using sparse matrix in the HYB format on the CPU.\n");

	indexType hyb_cols_per_row = compute_hyb_cols_per_row(csr_h);
	hyb = csr_to_hyb<indexType, valueType>(csr_h, hyb_cols_per_row, 1);
	
	x = new_host_array<valueType>(hyb.num_rows);
	
	b = copy_array_on_host<valueType>(b_h, hyb.num_rows);
}

template <typename indexType, typename valueType>
SysCPUHYB<indexType, valueType>::~SysCPUHYB()
{
	delete_host_array(x);
	delete_host_array(b);
	for (int i=0; i < vecArrayLength; ++i) delete_host_array(vecArray[i]);
	delete_host_matrix(hyb);
	free(vecArray);
}

template <typename indexType, typename valueType>
valueType ** SysCPUHYB<indexType,  valueType>::initVectors(valueType **x_e, valueType **b_e, const int length)
{
	*b_e = b;
	*x_e = x;
	vecArrayLength = length;
	vecArray = (valueType **) malloc(vecArrayLength * sizeof(valueType*));
	for (int i=0; i < vecArrayLength; ++i){
		vecArray[i] = new_host_array<valueType>(hyb.num_cols);
	}
	return vecArray;
}

template <typename indexType, typename valueType>
valueType SysCPUHYB<indexType,  valueType>::vectorVectorProduct(const valueType *v1, const valueType *v2) const
{
	valueType sum = 0.0;
	for (int i=0; i < hyb.num_rows; ++i){
		sum += v1[i] * v2[i];
	}
	return sum;
}

template <typename indexType, typename valueType>
void SysCPUHYB<indexType,  valueType>::matrixVectorProduct(const valueType *v, valueType *result) const
{
	/* the ELL part */
    for (indexType row = 0; row < hyb.ell.num_rows; row++){
		valueType sum = (valueType) 0.0;
	
		for(indexType n = 0; n < hyb.ell.num_cols_per_row; n++){
		    const valueType A_ij = hyb.ell.Ax[row + hyb.ell.stride * n];
		    if (A_ij != 0){
		        const indexType col = hyb.ell.Aj[row + hyb.ell.stride * n];
		        sum += A_ij * v[col];
		    }
		}
		result[row] = sum;
	}
	/* the COO part */
    for(indexType n = 0; n < hyb.coo.num_nonzeros; n++){
        result[hyb.coo.I[n]] += hyb.coo.V[n] * v[hyb.coo.J[n]];
    }
}

template <typename indexType, typename valueType>
void SysCPUHYB<indexType,  valueType>::axpy(const valueType a, const valueType *x_input, valueType *y) const
{
	for (int i=0; i < hyb.num_rows; i++){
		y[i] = a*x_input[i] + y[i];
	}
}

template <typename indexType, typename valueType>
void SysCPUHYB<indexType,  valueType>::scale(const valueType scalar, valueType *v) const
{
	if (scalar == (valueType) 0.0){
		for (int i=0; i < hyb.num_rows; i++) v[i] = 0.0;
	} else{
		for (int i=0; i < hyb.num_rows; i++){
			v[i] *= scalar;
		}
	}
}

template <typename indexType, typename valueType>
void SysCPUHYB<indexType,  valueType>::getSolution(valueType **solution) const
{
	for (int i=0; i < hyb.num_rows; i++){
		(*solution)[i] = x[i];
	}
}

template <typename indexType, typename valueType>
void SysCPUHYB<indexType,  valueType>::printVector(const valueType *v, const int numbersPerLine) const
{
	for (unsigned int line=0; line < hyb.num_rows/numbersPerLine; line++){
		for (unsigned int row=0; row< numbersPerLine; row++){
			writeToScreen1("%g ", v[line*numbersPerLine + row]);
		}
		writeToScreen("\n");
	}
	writeToScreen("\n");
}

template <typename indexType, typename valueType>
void SysCPUHYB<indexType, valueType>::printMatrix() const
{
	writeToScreen("ELLPACK part:\n");
	for (indexType i = 0; i < hyb.ell.stride; ++i){
		for (indexType k = 0; k < hyb.ell.num_cols_per_row; ++k){
			writeToScreen1("%g\t", hyb.ell.Ax[i + hyb.ell.stride * k]);
		}
		writeToScreen("\n");
	}
	writeToScreen("COO part:\n");
	for (indexType i = 0; i < hyb.coo.num_nonzeros; ++i){
		writeToScreen1("%i\t", (int) hyb.coo.I[i]);
		writeToScreen1("%i\t", (int) hyb.coo.J[i]);
		writeToScreen1("%g\n", hyb.coo.V[i]);
	}
}

#endif /*SysCPUHYB_H_*/
