#ifndef SYSCUBLAS_H_
#define SYSCUBLAS_H_
#include "CGsystem.h"
#include "cublas.h"
#include "commonHeader.h"
#include "cutil_inline.h"

template <class T>
class SysCublas : public CGsystem<T>
{
public:
	SysCublas(const T *A_h, const T *b_h, const size_type numOfElements);
	virtual ~SysCublas();
	virtual T ** initVectors(T **x, T **b, const int length);
	virtual T vectorVectorProduct(const T *v1, const T *v2) const;
	virtual void matrixVectorProduct(const T *v, T *result) const;
	virtual void axpy(const T a, const T *x, T *y) const;
	virtual void scale(const T scalar, T *v) const;
	virtual void copy(const T *src, T **dst) const {
			if (sizeof(T) == sizeof(double)){
				cublasDcopy(n, (double *)src, 1, (double *)*dst, 1);
			} else {
				cublasScopy(n, (float *)src, 1, (float *)*dst, 1);
			}
			CUBLAS_SAFE_CALL(cublasGetError(), (char*) "copy");
	}
	virtual void getSolution(T **solution) const;
	virtual size_type getN() const {return n;}
	virtual void printVector(const T *v, const int numbersPerLine) const;
	virtual void printMatrix() const {}
private:
	//host variables:
	cublasStatus status;
	size_type n;
	int vecArrayLength;
	unsigned int availableMemory;
	//device variables:
	T *A;
	T *x, *b;
	T **vecArray;
	//functions
	void CUBLAS_SAFE_CALL(cublasStatus status) const;
	void CUBLAS_SAFE_CALL(cublasStatus status, char* reason) const;
};


/*Create the system Ax = b to solve conjugate gradient with use of CUBLAS.
 * @param	A	The matrix A, as imported from MATLAB
 * 					It is stored in MATLAB format as the transposed of C standard, 
 * 					but CUBLAS seems to use the same convention.
 */
template <class T>
SysCublas<T>::SysCublas(const T *A_h, const T *b_h, const size_type dim) : n(dim)
{
	writeToScreen("Using full matrix in CUBLAS.\n");

	if (cuMemGetInfo(&availableMemory, NULL) != CUDA_SUCCESS) printf("Error reading memory status\n");
	if (n*n*(sizeof(T) + 10) > availableMemory){
		printf("Not enough memory, available memory is: \t%u\n\n", availableMemory);
		writeError("Not enough memory\n");
	}
	
	/* Allocat space for the matrix A */
	CUBLAS_SAFE_CALL(cublasAlloc(n*n, sizeof(T), (void**)&A), (char*) "initialize 1");
	CUBLAS_SAFE_CALL(cublasAlloc(n, sizeof(T), (void**)&b), (char*) "initialize 1");
	
	/* Allocat space for vectors */
	CUBLAS_SAFE_CALL(cublasAlloc(n, sizeof(T), (void**)&x), (char*) "initialize5");
	/* Initialize the matrix A on the GPU */
	CUBLAS_SAFE_CALL(cublasSetVector(n*n, sizeof(T), (void*)A_h, 1, (void*) A, 1), (char*) "initialize2");
	/* Initialize the vector b */
	CUBLAS_SAFE_CALL(cublasSetVector(n, sizeof(T), (void*)b_h, 1, (void*) b, 1), (char*) "initialize3");
}

template <class T>
SysCublas<T>::~SysCublas()
{
	CUBLAS_SAFE_CALL(cublasFree(x), (char*) "free");
	CUBLAS_SAFE_CALL(cublasFree(b));
	for (int i=0; i < vecArrayLength; ++i) CUBLAS_SAFE_CALL(cublasFree(vecArray[i]));
	CUBLAS_SAFE_CALL(cublasFree(A), (char*) "free");
}

template <typename T>
T ** SysCublas<T>::initVectors(T **x_e, T **b_e, const int length)
{
	*b_e = b;
	*x_e = x;
	vecArrayLength = length;
	vecArray = (T **) malloc(vecArrayLength * sizeof(T*));
	for (int i=0; i < vecArrayLength; ++i){
		CUBLAS_SAFE_CALL(cublasAlloc(n, sizeof(T), (void**)&vecArray[i]), (char*) "initVectors");
	}
	return vecArray;
}

template <class T>
T SysCublas<T>::vectorVectorProduct(const T *v1, const T *v2) const{
	T result;
	if (sizeof(T) == sizeof(double)){
		result = cublasDdot (n, (double *)v1, 1, (double *)v2, 1);
	} else {
		result = cublasSdot (n, (float *) v1, 1, (float *)v2, 1);
	}
	CUBLAS_SAFE_CALL(cublasGetError(), (char*) "vectorVector");
	return result;
}

template <class T>
void SysCublas<T>::matrixVectorProduct(const T *v, T *y) const
{
	if (sizeof(T) == sizeof(double)){
		cublasDgemv('n', n, n,1, (double *)A, n, (double *)v, 1, 0, (double *)y, 1);
	} else {
		cublasSgemv('n', n, n,1, (float *)A, n, (float *)v, 1, 0, (float *)y, 1);
	}
	CUBLAS_SAFE_CALL(cublasGetError(), (char*) "matrixVector");
}
template <class T>
void SysCublas<T>::axpy(const T a, const T *x, T *y) const
{
	if (sizeof(T) == sizeof(double)){
		cublasDaxpy(n, a, (double *)x, 1, (double *)y, 1);
	} else {
		cublasSaxpy(n, a, (float *)x, 1, (float *)y, 1);
	}
	CUBLAS_SAFE_CALL(cublasGetError(), (char*) "axpy");
}

template <class T>
void SysCublas<T>::scale(const T scalar, T *v) const
{
	if (scalar == (T)0.0){
		T *zero = (T *) malloc(n * sizeof(T));
		for (int i=0; i< n; ++i) zero[i] = (T) 0.0;
		cublasSetVector(n, sizeof(T), zero, 1, v, 1);
		return;
	} /* else */
	if (sizeof(T) == sizeof(double)){
		cublasDscal(n, (double)scalar, (double *)v, 1);
	} else {
		cublasSscal(n, (float)scalar, (float *)v, 1);
	}
	CUBLAS_SAFE_CALL(cublasGetError(), (char*) "scale");
}

template <class T>
void SysCublas<T>::getSolution(T **solution) const 
{
	CUBLAS_SAFE_CALL(cublasGetVector(n, sizeof(T), x, 1, *solution, 1), (char*) "getVector");
}

template <class T>
void SysCublas<T>::printVector(const T *v, const int numbersPerLine) const
{
	/* Copy the vector back to host memory, then print it to the screen*/
	double *output = new double[n];
	CUBLAS_SAFE_CALL(cublasGetVector(n, sizeof(T), v, 1, output, 1), (char*) "printVector");
	for (unsigned int line=0; line < n/numbersPerLine; line++){
		for (unsigned int row=0; row< numbersPerLine; row++){
			writeToScreen1("%g ", output[line*numbersPerLine + row]);
		}
		writeToScreen("\n");
	}
	delete[] output;
}

template <class T>
void SysCublas<T>::CUBLAS_SAFE_CALL(cublasStatus status) const {CUBLAS_SAFE_CALL(status, (char*)"");}
template <class T>
void SysCublas<T>::CUBLAS_SAFE_CALL(cublasStatus status, char* explanation) const{
	if (status != CUBLAS_STATUS_SUCCESS) writeToScreen1("%s", explanation);
	switch (status){
		case CUBLAS_STATUS_SUCCESS:
		return;
		case CUBLAS_STATUS_NOT_INITIALIZED: 
		writeError("CUBLAS library not initialized\n");
		case CUBLAS_STATUS_ALLOC_FAILED:
		writeError("resource allocation failed\n");
		case CUBLAS_STATUS_INVALID_VALUE:
		writeError("unsupported numerical value was passed to function\n");
		case CUBLAS_STATUS_ARCH_MISMATCH:
		writeError("function requires an architectural feature absent from the architecture of the device\n");
		case CUBLAS_STATUS_MAPPING_ERROR:
		writeError("access to GPU memory space failed\n");
		case CUBLAS_STATUS_EXECUTION_FAILED:
		writeError("GPU program failed to execute\n");
		case CUBLAS_STATUS_INTERNAL_ERROR:
		writeError("an internal CUBLAS operation failed\n");
		default:
		writeError("an undefined error occured\n");
	}
}

#endif /*SYSCUBLAS_H_*/
