#ifndef SysDIA_H_
#define SysDIA_H_
#include "CGsystem.h"
#include "commonHeader.h"
#include "spmv/sparse_formats.h"
#include "spmv/spmv_dia_device.cu.h"
#include <MyProject_kernel.cu>

#include "cutil_inline.h"
#include <cuda.h>

template <typename indexType, typename valueType, bool useCache>
class SysDIA : public CGsystem<valueType>
{
public:
	SysDIA(const csr_matrix<indexType, valueType> csr_h, const valueType *b, const int bs = 32, const int lpt = 4);
	virtual ~SysDIA();
	virtual valueType ** initVectors(valueType **x, valueType **b, const int length);
	virtual valueType vectorVectorProduct(const valueType*v1, const valueType *v2) const;
	virtual void matrixVectorProduct(const valueType *v, valueType *result) const;
	virtual void axpy(const valueType a, const valueType *x, valueType *y) const;
	virtual void scale(const valueType scalar, valueType *v) const;
	virtual void copy(const valueType *src, valueType **dst) const {cutilSafeCall(cudaMemcpy(*dst, src, dia.num_cols*sizeof(valueType), cudaMemcpyDeviceToDevice));}
	virtual void getSolution(valueType **solution) const;
	virtual size_type getN() const {return dia.num_cols;};
	virtual void printVector(const valueType *v, const int numbersPerLine) const;
	virtual void printMatrix() const;
private:
	//host variables:
	const unsigned int BLOCK_SIZE;
	const unsigned int LOADS_PER_THREAD;
	unsigned int availableMemory;
	int vecArrayLength;
	//hybrid variable:
	dia_matrix<indexType, valueType> dia;
	//device variables:
	valueType *x, *b;
	valueType **vecArray;
};

template <typename indexType, typename valueType, bool useCache>
SysDIA<indexType,  valueType, useCache>::SysDIA(const csr_matrix<indexType, valueType> csr_h, const valueType *b_h, const int bs, const int lpt)
	: BLOCK_SIZE(bs), LOADS_PER_THREAD(lpt)
{
	writeToScreen1("Using %s sparse matrix in the DIA format ", (sizeof(valueType) == sizeof(float)) ? "single precision" : "");
	writeToScreen1("with%s texturefetching.\n", (useCache) ? "": "out" );
	
	if (cuMemGetInfo(&availableMemory, NULL) != CUDA_SUCCESS) printf("Error reading memory status\n");
	unsigned int maxDiags = availableMemory / (csr_h.num_rows * sizeof(valueType) );
	maxDiags -= 10;
	dia_matrix<indexType, valueType> dia_h = csr_to_dia<indexType, valueType>(csr_h, maxDiags);
	if (dia_h.num_nonzeros == 0){
		printf("Not enough memory, available memory is: \t%u\n\n", availableMemory);
		writeError("Not enough memory\n");
	}
	if (dia_h.num_diags > 512) writeError("Too many diagonals\n");

	dia = copy_matrix_to_device(dia_h);
	
	x = new_device_array<valueType>(dia_h.num_cols);
	b = new_device_array<valueType>(dia_h.num_cols);

	memcpy_to_device<valueType>(b, b_h, dia_h.num_cols);
	
	delete_host_matrix(dia_h);
	cutilSafeCall(cudaGetLastError());
}

template <typename indexType, typename valueType, bool useCache>
SysDIA<indexType,  valueType, useCache>::~SysDIA()
{
	delete_device_array(x);
	delete_device_array(b);
	for (int i=0; i < vecArrayLength; ++i) delete_device_array(vecArray[i]);
	delete_device_matrix(dia);
	free(vecArray);
}

template <typename indexType, typename valueType, bool useCache>
valueType ** SysDIA<indexType,  valueType, useCache>::initVectors(valueType **x_e, valueType **b_e, const int length)
{
	*b_e = b;
	*x_e = x;
	vecArrayLength = length;
	vecArray = (valueType **) malloc(vecArrayLength * sizeof(valueType*));
	for (int i=0; i < vecArrayLength; ++i){
		vecArray[i] = new_device_array<valueType>(dia.num_cols);
	}
	return vecArray;
}

template <typename indexType, typename valueType, bool useCache>
valueType SysDIA<indexType,  valueType, useCache>::vectorVectorProduct(const valueType *v1, const valueType *v2) const
{
	valueType *result_d, result_h;
	//we prefer the number of blocks to be smaller than log2(n).
	const dim3 blockSize(BLOCK_SIZE);
    const dim3 gridSize((dia.num_rows-1) / (blockSize.x * LOADS_PER_THREAD) + 1);
	cutilSafeCall(cudaMalloc((void**) &result_d, gridSize.x * sizeof(valueType)));
	if (useCache){
		bind_x(v1);
		bind_y(v2);
	}
	switch (BLOCK_SIZE){
		case 1:		vectorVector_kernel<1, valueType, useCache>   <<< gridSize, blockSize,    sizeof(valueType) >>>(v1, v2, dia.num_rows, result_d);	break;
		case 2:		vectorVector_kernel<2, valueType, useCache>   <<< gridSize, blockSize,    2*sizeof(valueType) >>>(v1, v2, dia.num_rows, result_d);	break;
		case 4:		vectorVector_kernel<4, valueType, useCache>   <<< gridSize, blockSize,    4*sizeof(valueType) >>>(v1, v2, dia.num_rows, result_d);	break;
		case 8:		vectorVector_kernel<8, valueType, useCache>   <<< gridSize, blockSize,    8*sizeof(valueType) >>>(v1, v2, dia.num_rows, result_d);	break;
		case 16:		vectorVector_kernel<16, valueType, useCache>  <<< gridSize, blockSize,   16*sizeof(valueType) >>>(v1, v2, dia.num_rows, result_d);	break;
		case 32:		vectorVector_kernel<32, valueType, useCache>  <<< gridSize, blockSize,   32*sizeof(valueType) >>>(v1, v2, dia.num_rows, result_d);	break;
		case 64:		vectorVector_kernel<64, valueType, useCache>  <<< gridSize, blockSize,   64*sizeof(valueType) >>>(v1, v2, dia.num_rows, result_d);	break;
		case 128:	vectorVector_kernel<128, valueType, useCache> <<< gridSize, blockSize,  128*sizeof(valueType) >>>(v1, v2, dia.num_rows, result_d);	break;
		case 256:	vectorVector_kernel<256, valueType, useCache> <<< gridSize, blockSize,  256*sizeof(valueType) >>>(v1, v2, dia.num_rows, result_d);	break;
		case 512:	vectorVector_kernel<512, valueType, useCache> <<< gridSize, blockSize,  512*sizeof(valueType) >>>(v1, v2, dia.num_rows, result_d);	break;
	}
	if (useCache){
		unbind_y(v2);
		unbind_x(v1);
	}
	cutilSafeCall(cudaGetLastError());
	const dim3 one(1);
	//sum_kernel the results from each block. 
	//gridSize is less than 512,
	//so we only need one block with the size of the Grid for the last kernel.
	if (useCache) bind_x(result_d);
	switch (gridSize.x){
		case 1:	break;
		case 2:		sum_kernel<2, valueType, useCache>   <<< one, gridSize,   2*sizeof(valueType) >>>(result_d, gridSize.x, result_d);	break;
		case 4:		sum_kernel<4, valueType, useCache>   <<< one, gridSize,   4*sizeof(valueType) >>>(result_d, gridSize.x, result_d);	break;
		case 8:		sum_kernel<8, valueType, useCache>   <<< one, gridSize,   8*sizeof(valueType) >>>(result_d, gridSize.x, result_d);	break;
		case 16:		sum_kernel<16, valueType, useCache>  <<< one, gridSize,  16*sizeof(valueType) >>>(result_d, gridSize.x, result_d);	break;
		case 32:		sum_kernel<32, valueType, useCache>  <<< one, gridSize,  32*sizeof(valueType) >>>(result_d, gridSize.x, result_d);	break;
		case 64:		sum_kernel<64, valueType, useCache>  <<< one, gridSize,  64*sizeof(valueType) >>>(result_d, gridSize.x, result_d);	break;
		case 128:	sum_kernel<128, valueType, useCache> <<< one, gridSize, 128*sizeof(valueType) >>>(result_d, gridSize.x, result_d);	break;
		case 256:	sum_kernel<256, valueType, useCache> <<< one, gridSize, 256*sizeof(valueType) >>>(result_d, gridSize.x, result_d);	break;
		case 512:	sum_kernel<512, valueType, useCache> <<< one, gridSize, 512*sizeof(valueType) >>>(result_d, gridSize.x, result_d);	break;
	}
	if (useCache) unbind_x(result_d);
	cutilSafeCall(cudaGetLastError());
	//the first element of result_d holds the result.
	cutilSafeCall(cudaMemcpy(&result_h, result_d, sizeof(valueType), cudaMemcpyDeviceToHost));
	cutilSafeCall(cudaFree(result_d));
	return result_h;
}

template <typename indexType, typename valueType, bool useCache>
void SysDIA<indexType,  valueType, useCache>::matrixVectorProduct(const valueType *x_input, valueType *result) const
{
	scale(0.0, result);
	if (useCache)
		spmv_dia_tex_device<indexType, valueType>(dia, x_input, result);
	else
		spmv_dia_device<indexType, valueType>(dia, x_input, result);
	cutilSafeCall(cudaGetLastError());
}

template <typename indexType, typename valueType, bool useCache>
void SysDIA<indexType,  valueType, useCache>::axpy(const valueType a, const valueType *x_input, valueType *y) const
{
	const dim3 blockSize(BLOCK_SIZE);
	const dim3 gridSize((dia.num_rows-1) / BLOCK_SIZE + 1);
	if (useCache) bind_x(x_input);
	axpy_kernel<valueType, useCache> <<< gridSize, blockSize>>>(a, x_input, dia.num_rows, y);
	if (useCache) unbind_x(x_input);
	cutilSafeCall(cudaGetLastError());
}

template <typename indexType, typename valueType, bool useCache>
void SysDIA<indexType,  valueType, useCache>::scale(const valueType scalar, valueType *v) const
{
	const dim3 blockSize(BLOCK_SIZE);
	const dim3 gridSize(((dia.num_rows-1) / BLOCK_SIZE) + 1);
	if (scalar == (valueType)0.0)
		resetToZero_kernel<<< gridSize, blockSize >>>(dia.num_rows, v);
	else
		scale_kernel<<< gridSize, blockSize >>>(scalar, dia.num_rows, v);
	cutilSafeCall(cudaGetLastError());
}

template <typename indexType, typename valueType, bool useCache>
void SysDIA<indexType,  valueType, useCache>::getSolution(valueType **solution) const
{
	memcpy_to_host(*solution, x, dia.num_rows);
}

template <typename indexType, typename valueType, bool useCache>
void SysDIA<indexType,  valueType, useCache>::printVector(const valueType *v, const int numbersPerLine) const
{
	/* Copy the vector back to host memory, then print it to the screen if the vector lies on the GPU*/
	const valueType *output;
	/* Check if the variable v is a device variable assosiated with this object */
	if (v == x || v == b){
		valueType *temp = new valueType[dia.num_rows];
		memcpy_to_host<valueType>(temp, v, dia.num_rows);
		output = temp;
	} /* Otherwise, assume that the variable is a host-variable */
	else {
		output = v;
	}
	for (unsigned int line=0; line < dia.num_rows/numbersPerLine; line++){
		for (unsigned int row=0; row< numbersPerLine; row++){
			writeToScreen1("%g ", output[line*numbersPerLine + row]);
		}
		writeToScreen("\n");
	}
	writeToScreen("\n");
	if (output != v) delete[] output;
}

template <typename indexType, typename valueType, bool useCache>
void SysDIA<indexType, valueType, useCache>::printMatrix() const
{
	dia_matrix<indexType, valueType> dia_h;
    dia_h.diag_offsets = copy_array_to_device<int>(dia.diag_offsets, dia.num_diags);
	dia_h.diag_data = copy_array_to_host<valueType>(dia.diag_data, dia.stride*dia.num_diags);
	
	delete_host_matrix<indexType, valueType>(dia_h);
}

#endif /*SysDIA_H_*/
