#ifndef SysHYB_H_
#define SysHYB_H_
#include "CGsystem.h"
#include "commonHeader.h"
#include "spmv/sparse_formats.h"
#include "spmv/sparse_conversions.h"
#include "spmv/spmv_hyb_device.cu.h"
#include <MyProject_kernel.cu>

#include "cutil_inline.h"

template <typename indexType, typename valueType, bool useCache>
class SysHYB : public CGsystem<valueType>
{
public:
	SysHYB(const csr_matrix<indexType, valueType> csr_h, const valueType *b, const int bs = 32, const int lpt = 4);
	virtual ~SysHYB();
	virtual valueType ** initVectors(valueType **x, valueType **b, const int length);
	virtual valueType vectorVectorProduct(const valueType*v1, const valueType *v2) const;
	virtual void matrixVectorProduct(const valueType *v, valueType *result) const;
	virtual void axpy(const valueType a, const valueType *x, valueType *y) const;
	virtual void scale(const valueType scalar, valueType *v) const;
	virtual void copy(const valueType *src, valueType **dst) const {cutilSafeCall(cudaMemcpy(*dst, src, hyb.num_cols*sizeof(valueType), cudaMemcpyDeviceToDevice));}
	virtual void getSolution(valueType **solution) const;
	virtual size_type getN() const {return hyb.num_cols;};
	virtual void printVector(const valueType *v, const int numbersPerLine) const;
	virtual void printMatrix() const;
private:
	//host variables:
	const unsigned int BLOCK_SIZE;
	const unsigned int LOADS_PER_THREAD;
	unsigned int availableMemory;
	int vecArrayLength;
	//hybrid variable:
	hyb_matrix<indexType, valueType> hyb;
	//device variables:
	valueType *x, *b;
	valueType **vecArray;
};

template <typename indexType, typename valueType, bool useCache>
SysHYB<indexType,  valueType, useCache>::SysHYB(const csr_matrix<indexType, valueType> csr_h, const valueType *b_h, const int bs, const int lpt)
	: BLOCK_SIZE(bs), LOADS_PER_THREAD(lpt)
{
	writeToScreen1("Using%s sparse matrix in the HYB format ", (sizeof(valueType) == sizeof(float)) ? " single precision": "");
	writeToScreen1("with%s texturefetching.\n", (useCache) ? "": "out");
	
	hyb_matrix<indexType, valueType> hyb_h;
	
	indexType hyb_cols_per_row = compute_hyb_cols_per_row(csr_h);
	hyb_h = csr_to_hyb(csr_h, hyb_cols_per_row);

	if (cuMemGetInfo(&availableMemory, NULL) != CUDA_SUCCESS) printf("Error reading memory status\n");
	if (((hyb_h.ell.num_rows*hyb_cols_per_row*4)/3*(sizeof(valueType) + sizeof(indexType)) + (10*hyb_h.ell.num_rows*sizeof(valueType))) > availableMemory){
		printf("Available memory: \t%u\n", availableMemory);
		printf("Memory needed: \t%lu\n", ((hyb_h.ell.num_rows*hyb_cols_per_row*4)/3*(sizeof(valueType) + sizeof(indexType)) + (10*hyb_h.ell.num_rows*sizeof(valueType))));
		writeError("Not enough memory");
	}
	
	hyb = copy_matrix_to_device(hyb_h);

	x = new_device_array<valueType>(csr_h.num_cols);
	b = new_device_array<valueType>(csr_h.num_cols);
	
	memcpy_to_device<valueType>(b, b_h, csr_h.num_cols);
	delete_host_matrix(hyb_h);
}

template <typename indexType, typename valueType, bool useCache>
SysHYB<indexType,  valueType, useCache>::~SysHYB()
{
	delete_device_array(x);
	delete_device_array(b);
	for (int i=0; i < vecArrayLength; ++i) delete_device_array(vecArray[i]);
	delete_device_matrix(hyb);
}

template <typename indexType, typename valueType, bool useCache>
valueType ** SysHYB<indexType,  valueType, useCache>::initVectors(valueType **x_e, valueType **b_e, const int length)
{
	*b_e = b;
	*x_e = x;
	vecArrayLength = length;
	vecArray = (valueType **) malloc(vecArrayLength * sizeof(valueType*));
	for (int i=0; i < vecArrayLength; ++i){
		vecArray[i] = new_device_array<valueType>(hyb.num_cols);
	}
	return vecArray;
}

template <typename indexType, typename valueType, bool useCache>
valueType SysHYB<indexType,  valueType, useCache>::vectorVectorProduct(const valueType *v1, const valueType *v2) const
{
	valueType *result_d, result_h;
	//we prefer the number of blocks to be smaller than log2(n).
	const dim3 blockSize(BLOCK_SIZE);
    const dim3 gridSize((hyb.num_rows-1) / (blockSize.x * LOADS_PER_THREAD) + 1);
	cutilSafeCall(cudaMalloc((void**) &result_d, gridSize.x * sizeof(valueType)));
	if (useCache) {
		bind_x(v1);
		bind_y(v2);
	}
	switch (BLOCK_SIZE){
		case 1:		vectorVector_kernel<1, 	valueType, useCache>   <<< gridSize, blockSize,    sizeof(valueType) >>>(v1, v2, hyb.num_rows, result_d);	break;
		case 2:		vectorVector_kernel<2, 	valueType, useCache>   <<< gridSize, blockSize,    2*sizeof(valueType) >>>(v1, v2, hyb.num_rows, result_d);	break;
		case 4:		vectorVector_kernel<4, 	valueType, useCache>   <<< gridSize, blockSize,    4*sizeof(valueType) >>>(v1, v2, hyb.num_rows, result_d);	break;
		case 8:		vectorVector_kernel<8, 	valueType, useCache>   <<< gridSize, blockSize,    8*sizeof(valueType) >>>(v1, v2, hyb.num_rows, result_d);	break;
		case 16:		vectorVector_kernel<16,	valueType, useCache>  <<< gridSize, blockSize,   16*sizeof(valueType) >>>(v1, v2, hyb.num_rows, result_d);	break;
		case 32:		vectorVector_kernel<32,	valueType, useCache>  <<< gridSize, blockSize,   32*sizeof(valueType) >>>(v1, v2, hyb.num_rows, result_d);	break;
		case 64:		vectorVector_kernel<64,	valueType, useCache>  <<< gridSize, blockSize,   64*sizeof(valueType) >>>(v1, v2, hyb.num_rows, result_d);	break;
		case 128:	vectorVector_kernel<128,valueType, useCache> <<< gridSize, blockSize,  128*sizeof(valueType) >>>(v1, v2, hyb.num_rows, result_d);	break;
		case 256:	vectorVector_kernel<256,valueType, useCache> <<< gridSize, blockSize,  256*sizeof(valueType) >>>(v1, v2, hyb.num_rows, result_d);	break;
		case 512:	vectorVector_kernel<512,valueType, useCache> <<< gridSize, blockSize,  512*sizeof(valueType) >>>(v1, v2, hyb.num_rows, result_d);	break;
	}
	if (useCache){
		unbind_x(v1);
		unbind_y(v2);
	}
	cutilSafeCall(cudaGetLastError());
	const dim3 one(1);
	//sum_kernel the results from each block. 
	//gridSize is less than 512,
	//so we only need one block with the size of the Grid for the last kernel.
	if (useCache) bind_x(result_d);
	switch (gridSize.x){
		case 1:	break;
		case 2:		sum_kernel<2, 	valueType, useCache>   <<< one, gridSize,   2*sizeof(valueType) >>>(result_d, gridSize.x, result_d);	break;
		case 4:		sum_kernel<4, 	valueType, useCache>   <<< one, gridSize,   4*sizeof(valueType) >>>(result_d, gridSize.x, result_d);	break;
		case 8:		sum_kernel<8, 	valueType, useCache>   <<< one, gridSize,   8*sizeof(valueType) >>>(result_d, gridSize.x, result_d);	break;
		case 16:		sum_kernel<16, 	valueType, useCache>  <<< one, gridSize,  16*sizeof(valueType) >>>(result_d, gridSize.x, result_d);	break;
		case 32:		sum_kernel<32, 	valueType, useCache>  <<< one, gridSize,  32*sizeof(valueType) >>>(result_d, gridSize.x, result_d);	break;
		case 64:		sum_kernel<64, 	valueType, useCache>  <<< one, gridSize,  64*sizeof(valueType) >>>(result_d, gridSize.x, result_d);	break;
		case 128:	sum_kernel<128, valueType, useCache> <<< one, gridSize, 128*sizeof(valueType) >>>(result_d, gridSize.x, result_d);	break;
		case 256:	sum_kernel<256, valueType, useCache> <<< one, gridSize, 256*sizeof(valueType) >>>(result_d, gridSize.x, result_d);	break;
		case 512:	sum_kernel<512, valueType, useCache> <<< one, gridSize, 512*sizeof(valueType) >>>(result_d, gridSize.x, result_d);	break;
	}
	if (useCache) unbind_x(result_d);
	cutilSafeCall(cudaGetLastError());
	//the first element of result_d holds the result.
	cutilSafeCall(cudaMemcpy(&result_h, result_d, sizeof(valueType), cudaMemcpyDeviceToHost));
	cutilSafeCall(cudaFree(result_d));
	return result_h;
}

template <typename indexType, typename valueType, bool useCache>
void SysHYB<indexType,  valueType, useCache>::matrixVectorProduct(const valueType *v, valueType *result) const
{
	scale(0.0, result);
	if (useCache)
		spmv_hyb_tex_device<indexType, valueType>(hyb, v, result);
	else
		spmv_hyb_device<indexType, valueType>(hyb, v, result);
	cutilSafeCall(cudaGetLastError());
}

template <typename indexType, typename valueType, bool useCache>
void SysHYB<indexType,  valueType, useCache>::axpy(const valueType a, const valueType *x_input, valueType *y) const
{
	const dim3 blockSize(BLOCK_SIZE);
	const dim3 gridSize((hyb.num_rows-1) / BLOCK_SIZE + 1);
	if (useCache) bind_x(x_input);
	axpy_kernel<valueType, useCache> <<< gridSize, blockSize>>>(a, x_input, hyb.num_rows, y);
	if (useCache) unbind_x(x_input);
	cutilSafeCall(cudaGetLastError());
}

template <typename indexType, typename valueType, bool useCache>
void SysHYB<indexType,  valueType, useCache>::scale(const valueType scalar, valueType *v) const
{
	const dim3 blockSize(BLOCK_SIZE);
	const dim3 gridSize((hyb.num_rows-1) / BLOCK_SIZE + 1);
	if (scalar == (valueType)0.0)
		resetToZero_kernel<<< gridSize, blockSize >>>(hyb.num_rows, v);
	else
		scale_kernel<<< gridSize, blockSize >>>(scalar, hyb.num_rows, v);
	cutilSafeCall(cudaGetLastError());
}

template <typename indexType, typename valueType, bool useCache>
void SysHYB<indexType,  valueType, useCache>::getSolution(valueType **solution) const
{
	memcpy_to_host(*solution, x, hyb.num_rows);
}

template <typename indexType, typename valueType, bool useCache>
void SysHYB<indexType,  valueType, useCache>::printVector(const valueType *v, const int numbersPerLine) const
{
	/* Copy the vector back to host memory, then print it to the screen if the vector lies on the GPU*/
	const valueType *output;
	/* Check if the variable v is a device variable assosiated with this object */
	if (v == x || v == b){
		valueType *temp = new valueType[hyb.num_rows];
		memcpy_to_host<valueType>(temp, v, hyb.num_rows);
		output = temp;
	} /* Otherwise, assume that the variable is a host-variable */
	else {
		output = v;
	}
	for (unsigned int line=0; line < hyb.num_rows/numbersPerLine; line++){
		for (unsigned int row=0; row< numbersPerLine; row++){
			writeToScreen1("%g ", output[line*numbersPerLine + row]);
		}
		writeToScreen("\n");
	}
	writeToScreen("\n");
	if (output != v) delete[] output;
}

template <typename indexType, typename valueType, bool useCache>
void SysHYB<indexType, valueType, useCache>::printMatrix() const
{
	hyb_matrix<indexType, valueType> hyb_h;
	
	hyb_h.ell.Ax = copy_array_to_host<valueType>(hyb.ell.Ax, hyb.ell.stride * hyb.ell.num_cols_per_row);
	hyb_h.ell.Aj = copy_array_to_host<indexType>(hyb.ell.Aj, hyb.ell.stride * hyb.ell.num_cols_per_row);
	
	hyb_h.coo.V = copy_array_to_host<valueType>(hyb.coo.V, hyb.coo.num_nonzeros);
	hyb_h.coo.I = copy_array_to_host<indexType>(hyb.coo.I, hyb.coo.num_nonzeros);
	hyb_h.coo.J = copy_array_to_host<indexType>(hyb.coo.J, hyb.coo.num_nonzeros);
	
	writeToScreen("ELLPACK part:\n");
	for (indexType i = 0; i < hyb.ell.stride; ++i){
		for (indexType k = 0; k < hyb.ell.num_cols_per_row; ++k){
			writeToScreen1("%g\t", hyb_h.ell.Ax[i + hyb.ell.stride * k]);
		}
		writeToScreen("\n");
	}
	writeToScreen("COO part:\n");
	for (indexType i = 0; i < hyb.coo.num_nonzeros; ++i){
		writeToScreen1("%i\t", (int) hyb_h.coo.I[i]);
		writeToScreen1("%i\t", (int) hyb_h.coo.J[i]);
		writeToScreen1("%g\n", hyb_h.coo.V[i]);
	}
}

#endif /*SysHYB_H_*/
