#ifndef SysMyProject_H_
#define SysMyProject_H_
#include "CGsystem.h"

#include "cutil_inline.h"
#include <MyProject_kernel.cu>

template <class T>
class SysMyProject : CGsystem<T>
{
public:
	SysMyProject(const unsigned int nodim, const size_type *dim, const T *K, const T *b, size_t numberOfElements);	
	virtual ~SysMyProject();
	virtual T ** initVectors(T **x, T **b, const int length);
	virtual T vectorVectorProduct(const T *v1, const T *v2) const;
	virtual void matrixVectorProduct(const T *v, T *result) const;
	virtual void axpy(const T a, const T *x, T *y) const;
	virtual void scale(const T scalar, T *v) const;
	virtual void copy(const T *src, T **dst) const {cutilSafeCall(cudaMemcpy(*dst, src, n*sizeof(T), cudaMemcpyDeviceToDevice));}
	virtual void getSolution(T **solution) const;
	virtual size_t getN() const {return n;};
	virtual void printVector(const T *v, const int numbersPerLine) const;
	virtual void printMatrix() const;
private:
	//host variables:
	static const unsigned int BLOCK_SIZE = 32;
	static const unsigned int LOADS_PER_THREAD = 4;
	size_type n;
	unsigned int numberOfDiagonals;
	int vecArrayLength;
	//device variables:
	T *data;
	int *offset;
	T *x, *b;
	T **vecArray;
	//functions
	void prepareSystem(const T *data_h, const int *offset_h, const T *b);
};

/*Create the system Ax = b to solve conjugate gradient with operations from my project fall 2008.
 * @param	A	The matrix A, as imported from MATLAB
 * 					It is stored in MATLAB format as the transposed of C standard
 */
template <class T>
SysMyProject<T>::SysMyProject(const unsigned int nodim, const size_type *dim, const T *K, const T *b_h, size_type numberOfElements) : n(numberOfElements)
{
	writeToScreen("Using sparse matrix in the format used in my project (DIA).\n");
//	writeToScreen1("correct dimentions: %s\n", (n % (BLOCK_SIZE * LOADS_PER_THREAD) != 0) ? "no" : "yes");
	numberOfDiagonals = nodim*2 + 1;
	/* Build the diagoal representation of the matrix A */
	T hx, hy, hz;
	(nodim > 1) ? hx = (T)1.0 / dim[0] : hx = 1;
	(nodim > 1) ? hy = (T)1.0 / dim[1] : hy = 1;
	(nodim > 2) ? hz = (T)1.0 / dim[2] : hz = 1;
	numberOfDiagonals = nodim*2 + 1;
	T *data_h = new T[n*numberOfDiagonals];
	//create the spacing between the data.
	//The main diagonal has index 0, the next index 1 and -1, and the next after that is dim[0] and -dim[0] and so forth
	int *offset_h = new int[numberOfDiagonals];
	offset_h[nodim] = 0;
	for (unsigned int i=1; i<=nodim; ++i){
		offset_h[nodim + i] = 1;
		for (unsigned int j=0; j < i-1; j++){
			offset_h[nodim + i] *= dim[j];
		}
		offset_h[nodim -i] = -offset_h[nodim + i];
	}
	//zero out the diagonal vectors
	for (size_type i=0; i< n*numberOfDiagonals; ++i){
		data_h[i] = 0;
	}
	//build the matrix from the TPFA-scheme:
	T t_ij;
	for (size_type i=0; i< n; ++i){
		for (unsigned int k=0; k< numberOfDiagonals; ++k){
			//make j the second coordinate in the "real" matrix.
			if (offset_h[k] + (int)i >=0){
				unsigned int j = offset_h[k] + i;
				if (j < n){
					t_ij = 0;
					switch((int)k - (int)nodim){
						case 0:
							break;
						/* We have to check if the relation between i and j is suppose to contribute.
						 * Thas is if j is not representing a node on the boundary.*/
						//x-direction:
						case -1: // j is on the "negative" side of i, and we need to check if he is a edge node.
							if ((i % dim[0]) == 0) break;
							//K[q + 3*x + 3*Nx*y + 3*Nx*Ny*z] = qth-coordinate at point (x,y,z), q={x,y,z}={0,1,2}
							t_ij = (T)2*hz*hy/(hx*(1.0/K[0 + 3*i] + 1.0/(K[0 + 3*j])));
							break;
						case 1: //j is on the "positive" side of i, and we need to check if he is a edge node.
							if ((j % dim[0]) == 0) break;
							t_ij = (T)2*hz*hy/(hx*(1.0/K[0 + 3*i] + 1.0/(K[0 + 3*j])));
							break;
							
						//y-direction:
						case -2: // j is on the "negative" side of i, and we need to check if he is a edge node.
							if ((i % (dim[0]*dim[1])) < dim[0]) break;
							t_ij = (T)2*hx*hz/(hy*(1.0/K[1 + 3*i] + 1.0/(K[1 + 3*j])));
							break;
						case 2: //j is on the "positive" side of i, and we need to check if he is a edge node.
							if ((j % (dim[0]*dim[1])) < dim[0]) break;
							t_ij = (T)2*hx*hz/(hy*(1.0/K[1 + 3*i] + 1.0/(K[1 + 3*j])));
							break;
							
						//z-direction:
						case -3: // j is on the "negative" side of i, and we need to check if he is a edge node.
							if (i < (dim[0]*dim[1])) break;
							t_ij = (T)2*hx*hy/(hz*(1.0/K[2 + 3*i] + 1.0/(K[2 + 3*j])));
							break;
						case 3: //j is on the "positive" side of i, and we need to check if he is a edge node.
							if (j < (dim[0]*dim[1])) break;
							t_ij = (T)2*hx*hy/(hz*(1.0/K[2 + 3*i] + 1.0/(K[2 + 3*j])));
							break;
					}
					if (t_ij != 0){
						data_h[j*numberOfDiagonals + k] 				= -t_ij;
						data_h[i*numberOfDiagonals + numberOfDiagonals/2]	+= t_ij;
					}
				}
			}
		}
	}
	data_h[3] = data_h[3] + K[0] + K[1] + K[2];
	prepareSystem(data_h, offset_h, b_h);
	delete[] data_h;
	delete[] offset_h;
}

template <typename T>
T ** SysMyProject<T>::initVectors(T **x_e, T **b_e, const int length)
{
	*b_e = b;
	*x_e = x;
	vecArrayLength = length;
	vecArray = (T **) malloc(vecArrayLength * sizeof(T*));
	for (int i=0; i < vecArrayLength; ++i){
		cutilSafeCall(cudaMalloc((void**)&vecArray[i], n*sizeof(T)));
	}
	return vecArray;
}

template <class T>
void SysMyProject<T>::prepareSystem(const T *data_h, const int *offset_h, const T *b_h)
{
/*	char *nothing = "";
	cutilDeviceInit(1, &nothing); */

	cudaDeviceProp deviceProp;
	cudaGetDeviceProperties(&deviceProp, NULL);
//	if ((numberOfDiagonals*n*sizeof(T)) > deviceProp.totalGlobalMem) writeError("Not enough memory");
	
	cutilSafeCall(cudaMalloc((void**)&data, numberOfDiagonals*n*sizeof(T)));
	cutilSafeCall(cudaMalloc((void**)&offset, numberOfDiagonals*sizeof(int)));
	cutilSafeCall(cudaMalloc((void**)&x, n*sizeof(T)));
	cutilSafeCall(cudaMalloc((void**)&b, n*sizeof(T)));

	cutilSafeCall(cudaMemcpy(data, data_h, numberOfDiagonals*n*sizeof(T), cudaMemcpyHostToDevice));
	cutilSafeCall(cudaMemcpy(offset, offset_h, numberOfDiagonals*sizeof(int), cudaMemcpyHostToDevice));

	cutilSafeCall(cudaMemcpy(b, b_h, n*sizeof(T), cudaMemcpyHostToDevice));
	
	/*make x random for debug purposes
	T *temp = new T[n];
	for (int i=0; i<n; ++i){
		temp[i] = exp((T) rand()/RAND_MAX);
	}
	temp[2] = 1.0/0.0;//infinity
	temp[4] = 0.0/0.0;//nan
	cutilSafeCall(cudaMemcpy(x, temp, n*sizeof(T), cudaMemcpyHostToDevice));
	delete[] temp; */
}

template <class T>
SysMyProject<T>::~SysMyProject()
{
	cutilSafeCall(cudaFree(data));
	cutilSafeCall(cudaFree(offset));
	cutilSafeCall(cudaFree(x));
	cutilSafeCall(cudaFree(b));
	for (int i=0; i < vecArrayLength; ++i) cutilSafeCall(cudaFree(vecArray[i]));
	free(vecArray);
	/*char *nothing = "";
	cutilExit(1,&nothing);*/
}

template <class T>
T SysMyProject<T>::vectorVectorProduct(const T *v1, const T *v2) const
{
	T *result_d, result_h;
	//we prefer the number of blocks to be smaller than log2(n).
	dim3 blockSize(BLOCK_SIZE);
    dim3 gridSize((n-1) / (blockSize.x * LOADS_PER_THREAD) + 1);
	cutilSafeCall(cudaMalloc((void**) &result_d, gridSize.x * sizeof(T)));
	bind_x(v1);
	bind_y(v2);
	switch (BLOCK_SIZE){
		case 1:	break;
		case 2:		vectorVector_kernel<2, T, true>   <<< gridSize, blockSize,    2*sizeof(T) >>>(v1, v2, n, result_d);	break;
		case 4:		vectorVector_kernel<4, T, true>   <<< gridSize, blockSize,    4*sizeof(T) >>>(v1, v2, n, result_d);	break;
		case 8:		vectorVector_kernel<8, T, true>   <<< gridSize, blockSize,    8*sizeof(T) >>>(v1, v2, n, result_d);	break;
		case 16:		vectorVector_kernel<16, T, true>  <<< gridSize, blockSize,   16*sizeof(T) >>>(v1, v2, n, result_d);	break;
		case 32:		vectorVector_kernel<32, T, true>  <<< gridSize, blockSize,   32*sizeof(T) >>>(v1, v2, n, result_d);	break;
		case 64:		vectorVector_kernel<64, T, true>  <<< gridSize, blockSize,   64*sizeof(T) >>>(v1, v2, n, result_d);	break;
		case 128:	vectorVector_kernel<128, T, true> <<< gridSize, blockSize,  128*sizeof(T) >>>(v1, v2, n, result_d);	break;
		case 256:	vectorVector_kernel<256, T, true> <<< gridSize, blockSize,  256*sizeof(T) >>>(v1, v2, n, result_d);	break;
		case 512:	vectorVector_kernel<512, T, true> <<< gridSize, blockSize,  512*sizeof(T) >>>(v1, v2, n, result_d);	break;
	}
	cutilSafeCall(cudaGetLastError());
	dim3 one(1);
	//sum_kernel the results from each block. 
	//gridSize is less than 512,
	//so we only need one block with the size of the Grid for the last kernel.
	switch (gridSize.x){
		case 1:	break;
		case 2:		sum_kernel<2, T, false>   <<< one, gridSize,   2*sizeof(T) >>>(result_d, gridSize.x, result_d);	break;
		case 4:		sum_kernel<4, T, false>   <<< one, gridSize,   4*sizeof(T) >>>(result_d, gridSize.x, result_d);	break;
		case 8:		sum_kernel<8, T, false>   <<< one, gridSize,   8*sizeof(T) >>>(result_d, gridSize.x, result_d);	break;
		case 16:		sum_kernel<16, T, false>  <<< one, gridSize,  16*sizeof(T) >>>(result_d, gridSize.x, result_d);	break;
		case 32:		sum_kernel<32, T, false>  <<< one, gridSize,  32*sizeof(T) >>>(result_d, gridSize.x, result_d);	break;
		case 64:		sum_kernel<64, T, false>  <<< one, gridSize,  64*sizeof(T) >>>(result_d, gridSize.x, result_d);	break;
		case 128:	sum_kernel<128, T, false> <<< one, gridSize, 128*sizeof(T) >>>(result_d, gridSize.x, result_d);	break;
		case 256:	sum_kernel<256, T, false> <<< one, gridSize, 256*sizeof(T) >>>(result_d, gridSize.x, result_d);	break;
		case 512:	sum_kernel<512, T, false> <<< one, gridSize, 512*sizeof(T) >>>(result_d, gridSize.x, result_d);	break;
	}
	cutilSafeCall(cudaGetLastError());
	//the first element of result_d holds the result.
	cutilSafeCall(cudaMemcpy(&result_h, result_d, sizeof(T), cudaMemcpyDeviceToHost));
	cutilSafeCall(cudaFree(result_d));
	return result_h;
}

template <class T>
void SysMyProject<T>::matrixVectorProduct(const T *v, T *result) const 
{
	//Set up grid and thread dimensions
	dim3 blockSize(BLOCK_SIZE);
	dim3 gridSize((n-1) / BLOCK_SIZE + 1);

	//Execute the kernel
	switch (numberOfDiagonals){
		case 1: DM_matrixVector_kernel<T, 1><<< gridSize, blockSize, blockSize.x*sizeof(T) >>>(v, data, offset, n, result); break;
		case 3: DM_matrixVector_kernel<T, 3><<< gridSize, blockSize, blockSize.x*sizeof(T) >>>(v, data, offset, n, result); break;
		case 5: DM_matrixVector_kernel<T, 5><<< gridSize, blockSize, blockSize.x*sizeof(T) >>>(v, data, offset, n, result); break;
		case 7: DM_matrixVector_kernel<T, 7><<< gridSize, blockSize, blockSize.x*sizeof(T) >>>(v, data, offset, n, result); break;
	}
	cutilSafeCall(cudaGetLastError());
	cudaThreadSynchronize();
}

template <class T>
void SysMyProject<T>::axpy(const T a, const T *x, T *y) const
{
	dim3 blockSize(BLOCK_SIZE);
	dim3 gridSize((n-1) / BLOCK_SIZE + 1);
	axpy_kernel<T, false> <<< gridSize, blockSize>>>(a, x, n, y);
	cutilSafeCall(cudaGetLastError());
}

template <class T>
void SysMyProject<T>::scale(const T scalar, T *v) const
{
	dim3 blockSize(BLOCK_SIZE);
	dim3 gridSize((n-1) / BLOCK_SIZE + 1);
	scale_kernel<<< gridSize, blockSize >>>(scalar, n, v);
	cutilSafeCall(cudaGetLastError());
}

template <class T>
void SysMyProject<T>::getSolution(T **solution) const
{
	cutilSafeCall(cudaMemcpy(*solution, x, n*sizeof(T), cudaMemcpyDeviceToHost));
}

template <class T>
void SysMyProject<T>::printVector(const T *v, const int numbersPerLine) const
{
	/* Copy the vector back to host memory, then print it to the screen if the vector lies on the GPU*/
	const T *output;
	/* Check if the variable v is a device variable assosiated with this object */
	if (v == x || v ==b){
		T *temp = new T[n];
		cutilSafeCall(cudaMemcpy(temp, v, n*sizeof(T), cudaMemcpyDeviceToHost));
		output = temp;
	} /* Otherwise, assume that the variable is a host-variable */
	else {
		output = v;
	}
	for (unsigned int line=0; line < n/numbersPerLine; line++){
		for (unsigned int row=0; row< numbersPerLine; row++){
			writeToScreen1("%g ", output[line*numbersPerLine + row]);
		}
		writeToScreen("\n");
	}
	writeToScreen("\n");
	if (output != v) delete[] output;
}

template <class T>
void SysMyProject<T>::printMatrix() const
{
	/* Copy the matrix back to host memory and print it's diagonal elements to the screen */
	for (int line = 0; line < n; line++){
		for (int row = 0; row < numberOfDiagonals; row++){
			writeToScreen1("%g\t", data[line*numberOfDiagonals + row]);
		}
		writeToScreen("\n");
	}
}

#endif /*SysMyProject_H_*/
