#include "mat.h"
#include <string>
#include "spmv/sparse_formats.h"
#include "CGsystem.h"
#include "cg.h"
#include "rcg.h"
#include "SysCOO.cu.h"
#include "SysCSR.cu.h"
#include "SysCPUCSR.cu.h"
#include "SysCPUHYB.cu.h"
#include "SysELL.cu.h"
#include "SysDIA.cu.h"
#include "SysHYB.cu.h"
#include "SysCublas.h"

inline int log2(register unsigned int x){
	register unsigned int l=0;
	if (x >=  (1<<16)){ x >>= 16; l |= 16; }
	if (x >=  (1<<8)){ x >>= 8; l |= 8; }
	if (x >=  (1<<4)){ x >>= 4; l |= 4; }
	if (x >=  (1<<2)){ x >>= 2; l |= 2; }
	if (x >=  (1<<1)) l |= 1;
	return l;
}

int readFile(int argc, char **argv){
	if (argc < 2){
		printf("Usage: mexMaster <inputfile> <outputfile> [sparse-format]");
		printf(" where <inputfile> is the name of the MAT-file");
		printf(" containing A, b and tol to be read\n");
		printf("<outputfile> is where the solutionvector x ");
		printf(" and the calculated normed difference norm is stored.\n");
		printf("Sparseformat	 is a three-letter description of the sparse format\n");
		printf("\t COO = sparse coordinates format\n");
		printf("\t CSR = compressed sparse row format\n");
		printf("\t ELL = ELLPACK/ITPACK format\n");
		printf("\t HYB = a hybrid of ELL and COO\n");
		return 0;
	}
	
	if (cuInit(0) != CUDA_SUCCESS){
		printf("Error initializing CUDA\n");
		return 1;
	}
	CUdevice dev;
	CUcontext ctx;
	if (cuDeviceGet(&dev,0) != CUDA_SUCCESS){
		printf("Error getting device 0\n");
		return 1;
	}
	if (cuCtxCreate(&ctx, 0, dev) != CUDA_SUCCESS){
		printf("Error creating context for device 0\n");
		return 1;
	}
	if (ctx == NULL) printf("Warning: ctx is NULL\n");
	
	unsigned int memoryBefore, memoryAfter;
	int memStatus = cuMemGetInfo(&memoryBefore, NULL);
	switch (memStatus){
		case CUDA_SUCCESS:  break;
		case CUDA_ERROR_DEINITIALIZED:  		printf("CUDA_ERROR_DEINITIALIZED\n"); return memStatus;
		case CUDA_ERROR_NOT_INITIALIZED: 	printf("CUDA_ERROR_NOT_INITIALIZED\n"); return memStatus;
		case CUDA_ERROR_INVALID_CONTEXT:	printf("CUDA_ERROR_INVALID_CONTEXT\n"); return memStatus;
		case CUDA_ERROR_INVALID_VALUE: 		printf("CUDA_ERROR_INVALID_VALUE\n"); return memStatus;
	}
	printf("Available memory before launch: %u\n", memoryBefore);
		
 	printf("Reading file %s...\n\n", argv[1]);
 	MATFile *matFile = matOpen(argv[1], "r");
 	if (matFile == NULL){
 		printf("Error opening file %s\n", argv[1]);
 		return 1;
 	}
 	
 	mxArray *mxA = matGetVariable(matFile, "A");
 	mxArray *mxb = matGetVariable(matFile, "b");
 	mxArray *mxtol = matGetVariable(matFile, "tol");
 	if (mxA == NULL || mxb == NULL || mxtol == NULL){
 		printf("Unable to fetch A, b or tol from the file\n");
 		return 1;
 	}
 	
	/* check the number of dimensions of the matrix */
	if (mxGetN(mxb) != 1){
		printf("The second argument must be a column vector.\n");
		return 1;
	}
	mwSize n = mxGetM(mxb);
	if (mxGetM(mxA) != n || mxGetN(mxA) != n){
		printf("The size of the matrix must agree with the vector.\n");
		return 1;
	}
	if (!mxIsNumeric(mxtol)){
		printf("The tolerance has to be a scalar value.\n");
		return 1;
	}
	
	/* dimensions are correct, let's build the system */
	double rnorm;
	
	/* we need to figure out the ideal block-size and numbers of loads per thread (lpt)
	* n is the number of elements in a vector,
	* so we prefer lpt to be larger than log_2(n) */
	unsigned int lpt = log2((unsigned int) n)*2;
	unsigned int threads = n / lpt;
	unsigned int bs = (1 << log2(threads));
	if (bs > 512){
		bs = bs / 512;	//gs
		bs = 1 << (log2(bs) - 1);
		lpt = (n / (bs*512)) + 1;
		bs = 512;
	}
	writeToScreen1("n: %i\n", (int)n);
	writeToScreen1("blockSize: %i\n", (int)bs);
	writeToScreen1("loadsPerThread: %i\n", (int)lpt); 
    const dim3 gridSize((n-1) / (bs * lpt) + 1);
	writeToScreen1("gridSize: %i\n", (int)gridSize.x);
	
 	if (matClose(matFile) != 0){
 		printf("Error closing file\n");
 		return 1;
 	}
 	
 	MATFile *solutionFile = matOpen(argv[2], "w");
 	if (solutionFile == NULL){
 		printf("Unable to open solution file %s\n", argv[2]);
 		return 1;
 	}
 	
	const bool useCache = false;
	if (mxIsSparse(mxA) && mxIsDouble(mxb)){
		double *tol = (double *) mxGetData(mxtol);
		double *b = (double *) mxGetData(mxb);
	 	mxArray *mxSolution = mxCreateDoubleMatrix(n,1, mxREAL);
	 	double *solution = (double*)mxGetData(mxSolution);

		csr_matrix<size_t, double> csr;
		csr.Ax = mxGetPr(mxA);
		csr.Aj = mxGetIr(mxA);
		csr.Ap = mxGetJc(mxA);
		csr.num_cols = mxGetM(mxA);
		csr.num_rows = csr.num_cols;
		csr.num_nonzeros = csr.Ap[csr.num_rows];
		CGsystem<double> *system;
		CGsystem<float> *mlsystem;
		
		//string telling the sparse type to convert to
		std::string st;
		if (argc > 3) st = argv[3];
		else st = "HYB";
		if (st == "HYB" || st == "hyb") 				system = new SysHYB<mwSize, double, useCache>(csr, b, bs, lpt);
		else if (st == "COO" || st == "coo")		system = new SysCOO<mwSize, double, useCache>(csr, b, bs, lpt);
		else	if (st == "CSR" || st == "csr")			system = new SysCSR<mwSize, double, useCache>(csr, b, bs, lpt);
		else	if (st == "ELL" || st == "ell")			system = new SysELL<mwSize, double, useCache>(csr, b, bs, lpt);
		else	if (st == "DIA" || st == "dia")			system = new SysDIA<mwSize, double, useCache>(csr, b, bs, lpt);
		else if (st == "MLHYB" || st == "mlhyb")	{
			system = new SysHYB<size_type, double, true>(csr, b, bs, lpt);

			// converting csr and b to single precision:
			csr_matrix<size_type, float> csr_f;
			csr_f.num_rows = csr.num_rows;
			csr_f.num_cols = csr.num_cols;
			csr_f.num_nonzeros = csr.num_nonzeros;
			csr_f.Ap = copy_array_on_host(csr.Ap, csr.num_rows + 1);
			csr_f.Aj = copy_array_on_host(csr.Aj, csr.num_nonzeros);
			csr_f.Ax = new_host_array<float>(csr.num_nonzeros);
			for (int i=0; i < csr.num_nonzeros; ++i)		csr_f.Ax[i] = csr.Ax[i];

			float *b_f = (float*) malloc(csr.num_rows * sizeof(float));
			for (int i=0; i<csr.num_rows; i++){
				b_f[i] = (float) b[i];
			}
			mlsystem = new SysHYB<size_type, float, true>(csr_f, b_f, bs, lpt);
			free(b_f);
			delete_host_matrix(csr_f);
		}
		else if (st == "CPUCSR" || st == "cpucsr")	system = new SysCPUCSR<size_type, double>(csr, b);
		else if (st == "CPUHYB" || st == "cpuhyb")	system = new SysCPUHYB<size_type, double>(csr, b);
		else															system = new SysCSR<mwSize, double, useCache>(csr, b, bs, lpt); //default
		unsigned int maxIterations = n*4;
		if (mxGetM(mxtol) == 2 && (st == "MLHYB" || st == "mlhyb")){
			writeToScreen("solving in mixed precision\n");
			rnorm = rcg<double, float>(system, mlsystem, tol[0], tol[1], maxIterations, 0.8, bs);
			delete mlsystem;
		}
		else if (mxGetM(mxtol) == 2){
			writeToScreen("solving with restart\n");
			rnorm = rcg<double>(system, tol[0], tol[1], maxIterations, 0.8);
		}
		else if (mxGetM(mxtol) == 1)
			rnorm = cg<double>(system, tol[0], maxIterations);
		system->getSolution(&solution);
		printf("norm: %g\n", rnorm);
		delete system;

	 	printf("Writing to file %s...\n", argv[2]);
	 	int status = matPutVariable(solutionFile, "solution", mxSolution); 
	 	if (status != 0){
	 		printf("Error writing the solution vector to file\n");
	 		return 1;
	 	}
	} else if (mxIsSparse(mxA) && mxIsSingle(mxb)){
		float tol = mxGetScalar(mxtol);
		float *b = (float *) mxGetData(mxb);
		mwSize *dims = new mwSize[2];
		dims[0] = n;
		dims[1] = 1;
		mxArray *mxSolution = mxCreateNumericArray(2, dims, mxSINGLE_CLASS, mxREAL);
		delete[] dims;
	 	float *solution = (float*)mxGetData(mxSolution);

		csr_matrix<size_t, float> csr;
		csr.Aj = mxGetIr(mxA);
		csr.Ap = mxGetJc(mxA);
		csr.num_cols = mxGetM(mxA);
		csr.num_rows = csr.num_cols;
		csr.num_nonzeros = csr.Ap[csr.num_rows];
		csr.Ax = new float[csr.num_nonzeros];
		double *tempAx = (double *) mxGetData(mxA);
		for (int i=0; i<csr.num_nonzeros; i++){
			csr.Ax[i] = tempAx[i];
		}

		CGsystem<float> *system;
		
		//string telling the sparse type to convert to
		std::string st;
		if (argc > 3) st = argv[3];
		else st = "HYB";

		const bool useCache = true;
		if (st == "HYB" || st == "hyb") 				system = new SysHYB<mwSize, float, useCache>(csr, b, bs, lpt);
		else if (st == "COO" || st == "coo")		system = new SysCOO<mwSize, float, useCache>(csr, b, bs, lpt);
		else	if (st == "CSR" || st == "csr")			system = new SysCSR<mwSize, float, useCache>(csr, b, bs, lpt);
		else	if (st == "ELL" || st == "ell")			system = new SysELL<mwSize, float, useCache>(csr, b, bs, lpt);
		else	if (st == "DIA" || st == "dia")			system = new SysDIA<mwSize, float, useCache>(csr, b, bs, lpt);
		else if (st == "CPUCSR" || st == "cpucsr")	system = new SysCPUCSR<size_type, float>(csr, b);
		else if (st == "CPUHYB" || st == "cpuhyb")	system = new SysCPUHYB<size_type, float>(csr, b);
		else															system = new SysCSR<mwSize, float, useCache>(csr, b, bs, lpt); //default
		
		rnorm = cg<float>(system, tol, n);
		system->getSolution(&solution);
		printf("norm: %g\n", rnorm);
		delete system;
	 	printf("Writing to file %s...\n", argv[2]);
	 	int status = matPutVariable(solutionFile, "solution", mxSolution); 
	 	if (status != 0){
	 		printf("Error writing the solution vector to file\n");
	 		return 1;
	 	}
		delete[] tempAx;
	} else {
		double tol = (double) mxGetScalar(mxtol);
		double *b = (double *) mxGetData(mxb);
	 	mxArray *mxSolution = mxCreateDoubleMatrix(n,1, mxREAL);
	 	double *solution = (double*)mxGetData(mxSolution);
		double *A = (double *) mxGetData(mxA);
		SysCublas<double> system(A, b, n);
		rnorm = cg<double>((CGsystem<double>*)&system, tol, n);
		printf("norm: %g\n", rnorm);
		system.getSolution(&solution);
		
	 	printf("Writing to file %s...\n", argv[2]);
	 	int status = matPutVariable(solutionFile, "solution", mxSolution); 
	 	if (status != 0){
	 		printf("Error writing the solution vector to file\n");
	 		return 1;
	 	}
	}
 	
 	mxArray *mxNorm = mxCreateDoubleScalar(rnorm);
 	if (matPutVariable(solutionFile, "rnorm", mxCreateDoubleScalar(rnorm)) != 0){
 		printf("Error writing the residual norm to file\n");
 		return 1;
 	}
 	
 	printf("Solved the linear system successfully\n");
 	
	memStatus = cuMemGetInfo(&memoryAfter, NULL);
	switch (memStatus){
		case CUDA_SUCCESS:  break;
		case CUDA_ERROR_DEINITIALIZED:  		printf("CUDA_ERROR_DEINITIALIZED\n"); return memStatus;
		case CUDA_ERROR_NOT_INITIALIZED: 	printf("CUDA_ERROR_NOT_INITIALIZED\n"); return memStatus;
		case CUDA_ERROR_INVALID_CONTEXT:	printf("CUDA_ERROR_INVALID_CONTEXT\n"); return memStatus;
		case CUDA_ERROR_INVALID_VALUE: 		printf("CUDA_ERROR_INVALID_VALUE\n"); return memStatus;
	}
	
	printf("Available memory after execution: %u\n", memoryAfter);
	printf("Lost memory: %u\n", memoryBefore - memoryAfter);

	switch (cuCtxDetach(ctx) != CUDA_SUCCESS){
		case CUDA_SUCCESS:  break;
		case CUDA_ERROR_DEINITIALIZED:  		printf("CUDA_ERROR_DEINITIALIZED\n");
			printf("Error destroying context\n");
			return 1;
		case CUDA_ERROR_NOT_INITIALIZED: 	printf("CUDA_ERROR_NOT_INITIALIZED\n");
			printf("Error destroying context\n");
			return 1;
		case CUDA_ERROR_INVALID_CONTEXT:	printf("CUDA_ERROR_INVALID_CONTEXT\n");
			printf("Error destroying context\n");
			return 1;
		case CUDA_ERROR_INVALID_VALUE: 		printf("CUDA_ERROR_INVALID_VALUE\n");
			printf("Error destroying context\n");
			return 1;
	}

 	return 0;
 }

int main(int argc, char **argv)
{
	return readFile(argc, argv);
}