#ifndef MEXSTARTER_H_
#define MEXSTARTER_H_
#include "mex.h"
#include "SysCublas.h"
#include "SysMyProject.cu.h"
#include "SysCSR.cu.h"
#include "SysCPUCSR.cu.h"
#include "SysCPUHYB.cu.h"
#include "SysHYB.cu.h"
#include "SysCOO.cu.h"
#include "SysELL.cu.h"
#include "SysDIA.cu.h"
#include "cg.h"
#include "rcg.h"
#include <string>

inline int log2(register unsigned int x){
	register unsigned int l=0;
	if (x >=  (1<<16)){ x >>= 16; l |= 16; }
	if (x >=  (1<<8)){ x >>= 8; l |= 8; }
	if (x >=  (1<<4)){ x >>= 4; l |= 4; }
	if (x >=  (1<<2)){ x >>= 2; l |= 2; }
	if (x >=  (1<<1)) l |= 1;
	return l;
}

template <class T>
T importGrid(T *solution, const mxArray *prhs[])
{
	/* check the number of dimensions of the grid */
	if (mxGetN(prhs[0]) != 1 || mxGetN(prhs[2]) != 1)mexErrMsgIdAndTxt("MyToolbox:mexMaster:wrongDimension","First and third argument has to be column vectors.\n");
	mwSize d = mxGetM(prhs[0]);
	if (d > 3) mexErrMsgIdAndTxt("MyToolbox:mexMaster:wrongDimension","permeability-grids of dimension larger than three is currently unsupported.\n");
	mwSize n = mxGetM(prhs[2]);
	/* check if the dimensions of K corresponds to the dimensions given by the first vector: [Nx, Ny, Nz]'
	 * If Nz == 1, then K would be 3 x Nx x Ny x 1 = 3 x Nx x Ny. And similar if Ny also is 1. */
	size_type i; 
	for (i = d-1; (mxGetPr(prhs[0])[i] == 1); --i) ;
	if (mxGetNumberOfDimensions(prhs[1]) != i+2) mexErrMsgIdAndTxt("MyToolbox:mexMaster:wrongDimension", "The dimensional representation of K is not corresponding to the dimension vector.\nex: In 3D we have the input: mexMaster([Nx, Ny, Nz]', K, tol),\nwhere the dimension of K is 3 x Nx x Ny x Nz\n");
	while (i-->1){
		if ((int)mxGetDimensions(prhs[1])[i] != ((int*)mxGetData(prhs[0]))[i-1]) mexErrMsgIdAndTxt("MyToolbox:mexMaster:wrongDimension", "The dimensional representation of K is not corresponding to the dimension vector.\nex: In 3D we have the input: mexMaster([Nx, Ny, Nz]', K, tol),\nwhere the dimension of K is 3 x Nx x Ny x Nz\n");
	}
	if (*mxGetDimensions(prhs[1]) != d) mexErrMsgIdAndTxt("MyToolbox:mexMaster:wrongDimension","There are not the correct amount of values for K, corresponding to the dimensions of the problem\n");
	if (!mxIsNumeric(prhs[3])) mexErrMsgIdAndTxt("MyToolbox:mexMaster:wrongDimension","The tolerance has to be a scalar value.\n");

	/*dimensions are correct, let's initiate CUDA and build the system */
	/* Check if the first vector is of type double */
	if (!mxIsDouble(prhs[0])) mexErrMsgIdAndTxt("MyToolbox:mexMaster:prhs","The vector describing dimensions has to be of double precision.\n");
	
	
	size_type *dimVector = new size_type[d];
	double *temp = mxGetPr(prhs[0]);
	for (unsigned int i=0; i<d; i++){
		dimVector[i] = (size_type) temp[i];
	}
	T *K = (T*) mxGetData(prhs[1]);
	T *b = (T*) mxGetData(prhs[2]);
	T tol = (T) mxGetScalar(prhs[3]);
	T rnorm;

	SysMyProject<T> system(d, dimVector, K, b, n);
	rnorm = cg<T>((CGsystem<T>*)&system, tol, n);
	system.getSolution(&solution);
	delete[] dimVector;
	return rnorm;
}

template <class T>
T importMatrix(T *solution, const int nrhs, const mxArray *prhs[])
{
	/* check the number of dimensions of the matrix */
	if (mxGetN(prhs[1]) != 1) mexErrMsgIdAndTxt("MyToolbox:mexMaster:wrongDimension","The second argument must be a column vector.\n");
	size_t n = mxGetM(prhs[0]);
	if (mxGetM(prhs[0]) != n || mxGetN(prhs[0]) != n) mexErrMsgIdAndTxt("MyToolbox:mexMaster:wrongDimension","The size of the matrix must agree with the vector.\n");
	if (!mxIsNumeric(prhs[2])) mexErrMsgIdAndTxt("MyToolbox:mexMaster:wrongDimension","The tolerance has to be a scalar value.\n");
	
	/* dimensions are correct, let's initiate CUDA and build the system */

	if (cuInit(0) != CUDA_SUCCESS){
		writeToScreen("Error initializing CUDA\n");
	}
	CUdevice dev;
	CUcontext ctx;
	if (cuDeviceGet(&dev,0) != CUDA_SUCCESS){
		writeToScreen("Error getting device 0\n");
	}
	if (cuCtxAttach(&ctx, 0) != CUDA_SUCCESS){
		writeToScreen("Error attaching to current context, creating one\n");
		if (cuCtxCreate(&ctx, 0, dev) != CUDA_SUCCESS){
			writeToScreen("Error creating contexton device 0\n");
			return 0.0;
		}
	}
	if (ctx == NULL) writeToScreen("Warning: ctx is NULL\n");
	/*
	unsigned int memoryBefore, memoryAfter;
	int memStatus = cuMemGetInfo(&memoryBefore, NULL);
	switch (memStatus){
		case CUDA_SUCCESS:  break;
		case CUDA_ERROR_DEINITIALIZED:  		writeToScreen("CUDA_ERROR_DEINITIALIZED\n"); break;
		case CUDA_ERROR_NOT_INITIALIZED: 	writeToScreen("CUDA_ERROR_NOT_INITIALIZED\n"); break;
		case CUDA_ERROR_INVALID_CONTEXT:	writeToScreen("CUDA_ERROR_INVALID_CONTEXT\n"); break;
		case CUDA_ERROR_INVALID_VALUE: 		writeToScreen("CUDA_ERROR_INVALID_VALUE\n"); break;
	}
	writeToScreen1("Available memory before launch: %u\n", memoryBefore);
	*/
	/* Build the system */
	T rnorm;
	T *tol = (T*) mxGetData(prhs[2]);
	T *b = (T *) mxGetData(prhs[1]);
	int maxIterations;
	if (nrhs > 3){
		maxIterations =  *mxGetPr(prhs[4]);
		writeToScreen1("maxIterations = %i\n", maxIterations);
	}
	
	/* we need to figure out the ideal block-size and numbers of loads per thread (lpt)
	* n is the number of elements in a vector,
	* so we prefer lpt to be larger than log_2(n) */
	unsigned int lpt = log2((unsigned int) n)*2;
	unsigned int threads = n / lpt;
	unsigned int bs = (1 << log2(threads));
	if (bs > 512){
		bs = bs / 512;	//gs
		bs = 1 << (log2(bs) - 1);
		lpt = (n / (bs*512)) + 1;
		bs = 512;
	}
	writeToScreen1("n: %i\n", (int)n);
	writeToScreen1("blockSize: %i\n", (int)bs);
	writeToScreen1("loadsPerThread: %i\n", (int)lpt); 
    const dim3 gridSize((n-1) / (bs * lpt) + 1);
	writeToScreen1("gridSize: %i\n", (int)gridSize.x);

	
	if (mxIsSparse(prhs[0])){
		csr_matrix<mwIndex, T> csr;
		csr.num_cols = mxGetM(prhs[0]);
		csr.num_rows = csr.num_cols;
		csr.Aj = mxGetIr(prhs[0]);
		csr.Ap = mxGetJc(prhs[0]);
		csr.num_nonzeros = csr.Ap[csr.num_rows];
		if (sizeof(T) != sizeof(double)){
			csr.Ax = new T[csr.num_nonzeros];
			double *tempAx = (double *) mxGetData(prhs[0]);
			for (int i=0; i<csr.num_nonzeros; i++){
				csr.Ax[i] = (T) tempAx[i];
			}
		} else {
			csr.Ax = (T *) mxGetData(prhs[0]);
		}
		CGsystem<T> *system;
		CGsystem<float> *mlsystem;
		
		std::string st;
		if (nrhs == 5){
			//convert the last argument to a string to find the sparse type
//			const mxArray *string = prhs[3];
//			mwSize buflen = mxGetNumberOfElements(prhs[3]) + 1;
			mwSize buflen = (mxGetM(prhs[3]) * mxGetN(prhs[3]) * sizeof(mxChar)) + 1;
			char *sparseType = (char*) malloc(buflen * sizeof(char));
			 if (mxGetString(prhs[3], sparseType, buflen) != 0)
	 			mexErrMsgIdAndTxt("MyToolbox:mexMaster:nrhs","Could not convert string data from the last input.");
			st = sparseType;
	 		free(sparseType);
	 		if (true){
				if (st == "HYB" || st == "hyb")				system = new SysHYB<size_type, T, true>(csr, b, bs, lpt);
				else if (st == "COO" || st == "coo")	system = new SysCOO<size_type, T, true>(csr, b, bs, lpt);
				else if (st == "CSR" || st == "csr")		system = new SysCSR<size_type, T, true>(csr, b, bs, lpt);
				else if (st == "ELL" || st == "ell")		system = new SysELL<size_type, T, true>(csr, b, bs, lpt);
				else if (st == "DIA" || st == "dia")		system = new SysDIA<size_type, T, true>(csr, b, bs, lpt);
				else if (st == "MLHYB" || st == "mlhyb")	{
					system = new SysHYB<size_type, T, true>(csr, b, bs, lpt);
		
					// converting csr and b to single precision:
					csr_matrix<size_type, float> csr_f;
					csr_f.num_rows = csr.num_rows;
					csr_f.num_cols = csr.num_cols;
					csr_f.num_nonzeros = csr.num_nonzeros;
					csr_f.Ap = copy_array_on_host(csr.Ap, csr.num_rows + 1);
					csr_f.Aj = copy_array_on_host(csr.Aj, csr.num_nonzeros);
					csr_f.Ax = new_host_array<float>(csr.num_nonzeros);
					for (int i=0; i < csr.num_nonzeros; ++i)		csr_f.Ax[i] = csr.Ax[i];
		
					float *b_f = (float*) malloc(csr.num_rows * sizeof(float));
					for (int i=0; i<csr.num_rows; i++){
						b_f[i] = (float) b[i];
					}
					mlsystem = new SysHYB<size_type, float, true>(csr_f, b_f, bs, lpt);
					free(b_f);
					delete_host_matrix(csr_f);
				}
				else if (st == "CPUCSR" || st == "cpucsr")	system = new SysCPUCSR<size_type, T>(csr, b);
				else if (st == "CPUHYB" || st == "cpuhyb")	system = new SysCPUHYB<size_type, T>(csr, b);
				else														system = new SysHYB<size_type, T, true>(csr, b, bs, lpt);
	 		} else {
				if (st == "HYB" || st == "hyb")				system = new SysHYB<size_type, T, false>(csr, b, bs, lpt);
				else if (st == "COO" || st == "coo")	system = new SysCOO<size_type, T, false>(csr, b, bs, lpt);
				else if (st == "CSR" || st == "csr")		system = new SysCSR<size_type, T, false>(csr, b, bs, lpt);
				else if (st == "ELL" || st == "ell")		system = new SysELL<size_type, T, false>(csr, b, bs, lpt);
				else if (st == "DIA" || st == "dia")		system = new SysDIA<size_type, T, false>(csr, b, bs, lpt);
				else if (st == "CPUCSR" || st == "cpucsr")	system = new SysCPUCSR<size_type, T>(csr, b);
				else if (st == "CPUHYB" || st == "cpuhyb") 	system = new SysCPUHYB<size_type, T>(csr, b);
				else														system = new SysHYB<size_type, T, false>(csr, b, bs, lpt);
	 		}
		} //default is CSR
		else 														system = new SysCSR<size_type, T, true>(csr, b, bs, lpt);
		
		if ((st == "MLHYB" || st == "mlhyb")){
			if (mxGetM(prhs[2]) != 2) writeError("tolerance has to have two values when mixed precision is used"); 
			writeToScreen("solving in mixed precision\n");
			rnorm = rcg<T, float>(system, mlsystem, tol[0], tol[1], maxIterations, 0.8, bs);
			delete mlsystem;
		}
		else if (mxGetM(prhs[2]) == 2){
			writeToScreen("solving with restart\n");
			rnorm = rcg<T>(system, tol[0], tol[1], maxIterations, 0.8);
		}
		else if (mxGetM(prhs[2]) == 1){
			writeToScreen("solving the good-old fasion way\n");
			rnorm = cg<T>(system, tol[0], maxIterations);
		}
		system->getSolution(&solution);
		delete system;
		if (sizeof(T) != sizeof(double)) delete[] csr.Ax;
	} else {
		T *A = (T *) mxGetData(prhs[0]);
		SysCublas<T> system(A, b, n);
		if (mxGetM(prhs[2]) == 2){
			writeToScreen("solving with restart\n");
			rnorm = rcg<T>((CGsystem<T> *)&system, tol[0], tol[1], n, 0.8);
		}
		else if (mxGetM(prhs[2]) == 1){
			writeToScreen("solving the good-old fasion way\n");
			rnorm = cg<T>((CGsystem<T> *)&system, tol[0], n);
		}
		system.getSolution(&solution);
	}
	/*
	memStatus = cuMemGetInfo(&memoryAfter, NULL);
	switch (memStatus){
		case CUDA_SUCCESS:  break;
		case CUDA_ERROR_DEINITIALIZED:  		writeToScreen("CUDA_ERROR_DEINITIALIZED\n"); break;
		case CUDA_ERROR_NOT_INITIALIZED: 	writeToScreen("CUDA_ERROR_NOT_INITIALIZED\n"); break;
		case CUDA_ERROR_INVALID_CONTEXT:	writeToScreen("CUDA_ERROR_INVALID_CONTEXT\n"); break;
		case CUDA_ERROR_INVALID_VALUE: 		writeToScreen("CUDA_ERROR_INVALID_VALUE\n"); break;
	}
	
	writeToScreen1("Available memory after execution: %u\n", memoryAfter);
	writeToScreen1("Lost memory: %u\n", memoryBefore - memoryAfter);
	*/
	if (cuCtxDetach(ctx) != CUDA_SUCCESS)
		writeToScreen("Error detaching context\n");

	return rnorm;
}
#endif /*MEXSTARTER_H_*/
