/* Copyright 2008 NVIDIA Corporation.  All Rights Reserved */

#pragma once

#include "spmv/sparse_formats.h"
#include "spmv/texture.h"
#include "spmv/spmv_common_device.cu.h"

//////////////////////////////////////////////////////////////////////////////
// CSR SpMV kernels based on a vector model (one warp per row)
//////////////////////////////////////////////////////////////////////////////
//
// spmv_csr_vector_device
//   Each row of the CSR matrix is assigned to a warp.  The warp computes
//   y[i] = A[i,:] * x, i.e. the dot product of the i-th row of A with 
//   the x vector, in parallel.  This division of work implies that 
//   the CSR index and data arrays (Aj and Ax) are accessed in a contiguous
//   manner (but generally not aligned).  On GT200 these accesses are
//   coalesced, unlike kernels based on the one-row-per-thread division of 
//   work.  Since an entire 32-thread warp is assigned to each row, many 
//   threads will remain idle when their row contains a small number 
//   of elements.  This code relies on implicit synchronization among 
//   threads in a warp.
//
// spmv_csr_vector_tex_device
//   Same as spmv_csr_vector_tex_device, except that the texture cache is 
//   used for accessing the x vector.

template <typename IndexType, typename ValueType, unsigned int BLOCK_SIZE, bool UseCache>
__global__ void
spmv_csr_vector_kernel(const IndexType num_rows,
                       const IndexType * Ap, 
                       const IndexType * Aj, 
                       const ValueType * Ax, 
                       const ValueType * x, 
                             ValueType * y)
{
    __shared__ ValueType sdata[BLOCK_SIZE];
    __shared__ IndexType ptrs[BLOCK_SIZE/WARP_SIZE][2];
    
    const IndexType thread_id   = BLOCK_SIZE * blockIdx.x + threadIdx.x;  // global thread index
    const IndexType thread_lane = threadIdx.x & (WARP_SIZE-1);            // thread index within the warp
    const IndexType warp_id     = thread_id   / WARP_SIZE;                // global warp index
    const IndexType warp_lane   = threadIdx.x / WARP_SIZE;                // warp index within the CTA
    const IndexType num_warps   = (BLOCK_SIZE / WARP_SIZE) * gridDim.x;   // total number of active warps

    for(IndexType row = warp_id; row < num_rows; row += num_warps){
        // use two threads to fetch Ap[row] and Ap[row+1]
        // this is considerably faster than the more straightforward option
        if(thread_lane < 2)
            ptrs[warp_lane][thread_lane] = Ap[row + thread_lane];
        const IndexType row_start = ptrs[warp_lane][0]; //same as: row_start = Ap[row];
        const IndexType row_end   = ptrs[warp_lane][1]; //same as: row_end   = Ap[row+1];

        // compute local sum
        sdata[threadIdx.x] = 0;
        for(IndexType jj = row_start + thread_lane; jj < row_end; jj += WARP_SIZE)
            sdata[threadIdx.x] += Ax[jj] * fetch_x<UseCache>(Aj[jj], x);

        // reduce local sums to row sum (ASSUME: warpsize 32)
        if (thread_lane < 16) { sdata[threadIdx.x] += sdata[threadIdx.x + 16]; EMUSYNC; }
        if (thread_lane <  8) { sdata[threadIdx.x] += sdata[threadIdx.x +  8]; EMUSYNC; }
        if (thread_lane <  4) { sdata[threadIdx.x] += sdata[threadIdx.x +  4]; EMUSYNC; }
        if (thread_lane <  2) { sdata[threadIdx.x] += sdata[threadIdx.x +  2]; EMUSYNC; }
        if (thread_lane <  1) { sdata[threadIdx.x] += sdata[threadIdx.x +  1]; EMUSYNC; }

        // first thread writes warp result
        if (thread_lane == 0)
            y[row] += sdata[threadIdx.x];
    }
}

template <typename IndexType, typename ValueType>
void spmv_csr_vector_device(const csr_matrix<IndexType,ValueType>& d_csr, 
                            const ValueType * d_x, 
                                  ValueType * d_y)
{
    const unsigned int BLOCK_SIZE = 256;
    const unsigned int NUM_BLOCKS = MAX_THREADS/BLOCK_SIZE;
    
    spmv_csr_vector_kernel<IndexType, ValueType, BLOCK_SIZE, false> <<<NUM_BLOCKS, BLOCK_SIZE>>> 
        (d_csr.num_rows, d_csr.Ap, d_csr.Aj, d_csr.Ax, d_x, d_y);	

}

template <typename IndexType, typename ValueType>
void spmv_csr_vector_tex_device(const csr_matrix<IndexType,ValueType>& d_csr, 
                                const ValueType * d_x, 
                                      ValueType * d_y)
{
    const unsigned int BLOCK_SIZE = 256;
    const unsigned int NUM_BLOCKS = MAX_THREADS/BLOCK_SIZE;
    
    bind_x(d_x);
    
    spmv_csr_vector_kernel<IndexType,ValueType, BLOCK_SIZE, true> <<<NUM_BLOCKS, BLOCK_SIZE>>> 
        (d_csr.num_rows, d_csr.Ap, d_csr.Aj, d_csr.Ax, d_x, d_y);	

    unbind_x(d_x);
}

