/* Copyright 2008 NVIDIA Corporation.  All Rights Reserved */

#pragma once

#include "sparse_formats.h"
#include "utils.h"
#include "texture.h"

////////////////////////////////////////////////////////////////////////
// DIA SpMV kernels 
///////////////////////////////////////////////////////////////////////
//
// Diagonal matrices arise in grid-based discretizations using stencils.  
// For instance, the standard 5-point discretization of the two-dimensional 
// Laplacian operator has the stencil:
//      [  0  -1   0 ]
//      [ -1   4  -1 ]
//      [  0  -1   0 ]
// and the resulting DIA format has 5 diagonals.
//
// spmv_dia_device
//   Each thread computes y[i] += A[i,:] * x 
//   (the dot product of the i-th row of A with the x vector)
//
// spmv_dia_tex_device
//   Same as spmv_dia_device, except x is accessed via texture cache.
//


template <typename IndexType, typename ValueType, unsigned int BLOCK_SIZE, bool UseCache>
__global__ void
spmv_dia_kernel(const IndexType num_rows, 
                const IndexType num_cols, 
                const IndexType num_diags,
                const IndexType stride,
                const int       * diag_offsets,
                const ValueType * diag_data,
                const ValueType * x, 
                      ValueType * y)
{
    __shared__ int offsets[BLOCK_SIZE];

    if(threadIdx.x < num_diags)
        offsets[threadIdx.x] = diag_offsets[threadIdx.x];

    __syncthreads();

    const int row = large_grid_thread_id();

    if(row >= num_rows){ return; }

    ValueType sum = y[row];
    diag_data += row;

    for(IndexType n = 0; n < num_diags; n++){
        const int col = row + offsets[n];

        if(col >= 0 && col < num_cols){
            const ValueType A_ij = *diag_data;
            sum += A_ij * fetch_x<UseCache>(col, x);
        }

        diag_data += stride;
    }

    y[row] = sum;
}

template <typename IndexType, typename ValueType>
void spmv_dia_device(const dia_matrix<IndexType,ValueType>& d_dia, 
                     const ValueType * d_x, 
                           ValueType * d_y)
{
    const unsigned int BLOCK_SIZE = 512;
    const dim3 grid = make_large_grid(d_dia.num_rows, BLOCK_SIZE);
  
    assert(d_dia.num_diags < BLOCK_SIZE); // kernel doesn't handle larger numbers of diagonals

    spmv_dia_kernel<IndexType, ValueType, BLOCK_SIZE, false> <<<grid, BLOCK_SIZE>>>
        (d_dia.num_rows, d_dia.num_cols, d_dia.num_diags, d_dia.stride,
         d_dia.diag_offsets, d_dia.diag_data,
         d_x, d_y);
}

template <typename IndexType, typename ValueType>
void spmv_dia_tex_device(const dia_matrix<IndexType,ValueType>& d_dia, 
                         const ValueType * d_x, 
                               ValueType * d_y)
{
    const unsigned int BLOCK_SIZE = 512;
    const dim3 grid = make_large_grid(d_dia.num_rows, BLOCK_SIZE);
    
    assert(d_dia.num_diags < BLOCK_SIZE); // kernel doesn't handle larger numbers of diagonals
    
    bind_x(d_x);
   
    spmv_dia_kernel<IndexType, ValueType, BLOCK_SIZE, true> <<<grid, BLOCK_SIZE>>>
        (d_dia.num_rows, d_dia.num_cols, d_dia.num_diags, d_dia.stride,
         d_dia.diag_offsets, d_dia.diag_data,
         d_x, d_y);

    unbind_x(d_x);
}

