/* Copyright 2008 NVIDIA Corporation.  All Rights Reserved */

#pragma once

#include "cuda.h"


/*
 * These textures are (optionally) used to cache the 'x' vector in y += A*x
 * and x and y in r = y^T x;
 */
texture<float,1> tex_x_float;
texture<int2,1>  tex_x_double;
texture<float,1> tex_y_float;
texture<int2,1>  tex_y_double;

// Use int2 to pull doubles through texture cache
void bind_x(const float * x)
{   CUDA_SAFE_CALL(cudaBindTexture(NULL, tex_x_float, x));   }

void bind_x(const double * x)
{   CUDA_SAFE_CALL(cudaBindTexture(NULL, tex_x_double, x));   }

void bind_y(const float * y)
{   CUDA_SAFE_CALL(cudaBindTexture(NULL, tex_y_float, y));   }

void bind_y(const double * y)
{   CUDA_SAFE_CALL(cudaBindTexture(NULL, tex_y_double, y));   }

void unbind_x(const float * x)
{   CUDA_SAFE_CALL(cudaUnbindTexture(tex_x_float)); }
void unbind_x(const double * x)
{   CUDA_SAFE_CALL(cudaUnbindTexture(tex_x_double)); }
void unbind_y(const float * y)
{   CUDA_SAFE_CALL(cudaUnbindTexture(tex_y_float)); }
void unbind_y(const double * y)
{   CUDA_SAFE_CALL(cudaUnbindTexture(tex_y_double)); }
// Note: x and y is unused, but distinguishes the two functions

template <bool UseCache>
__inline__ __device__ float fetch_x(const int& i, const float * x)
{
    if (UseCache)
        return tex1Dfetch(tex_x_float, i);
    else
        return x[i];
}
template <bool UseCache>
__inline__ __device__ float fetch_y(const int& i, const float * y)
{
    if (UseCache)
        return tex1Dfetch(tex_y_float, i);
    else
        return y[i];
}

#if !defined(CUDA_NO_SM_13_DOUBLE_INTRINSICS)
template <bool UseCache>
__inline__ __device__ double fetch_x(const int& i, const double * x)
{
    if (UseCache){
        int2 v = tex1Dfetch(tex_x_double, i);
        return __hiloint2double(v.y, v.x);
    } else {
        return x[i];
    }
}
template <bool UseCache>
__inline__ __device__ double fetch_y(const int& i, const double * y)
{
    if (UseCache){
        int2 v = tex1Dfetch(tex_y_double, i);
        return __hiloint2double(v.y, v.x);
    } else {
        return y[i];
    }
}
#endif // !defined(CUDA_NO_SM_13_DOUBLE_INTRINSICS)

